"""
dbmechanic module for Pylons

Initial version by Isaac Csandl <nerkles _at_ gmail {d0t} com> based on the tg2 version.

"""
import os
import pkg_resources
import sqlalchemy
from tw.api import Widget
import pylons
from pylons.templating import cached_template, pylons_globals
from pylons.controllers import WSGIController
from pylons import c, request
from dbsprockets.sprockets import Sprockets
from dbsprockets.decorators import crudErrorCatcher, validate
from tw.api import WidgetBunch
from genshi.template import TemplateLoader


def render_genshi(template_name, genshi_loader, cache_key=None, cache_type=None,
                  cache_expire=None, fragment=False, format='xhtml'):
    """
    Copied out of Pylons so that we can specify the path to genshi templates.
    
    Render a template with Genshi
    
    Accepts the cache options ``cache_key``, ``cache_type``, and
    ``cache_expire`` in addition to fragment and format which are
    passed to Genshi's render function.
    
    """
    # Create a render callable for the cache function
    def render_template():
        # First, get the globals
        globs = pylons_globals()
        
        # Grab a template reference
        template = genshi_loader.load(template_name)
        
        return template.generate(**globs).render()
    
    return cached_template(template_name, render_template, cache_key=cache_key,
                           cache_type=cache_type, cache_expire=cache_expire,
                           ns_options=('fragment', 'format'),
                           fragment=fragment, format=format)


class DBMechanic(WSGIController):
    sprockets = None
    def __init__(self, provider, controller, *args, **kwargs):
        self.provider = provider
        self.sprockets = Sprockets(provider, controller)
        self.controller = controller
        #commonly used views
        c.w = WidgetBunch()
        sprocket = self.sprockets['databaseView']
        self.databaseValue = sprocket.session.getValue()
        self.databaseView  = sprocket.view.widget
        print 'controller:', self.controller
        
        self.databaseDict  = dict(controller=self.controller)
        self.genshi_loader = TemplateLoader([pkg_resources.resource_filename('dbsprockets.dbmechanic.frameworks.pylons', 'templates')])
        WSGIController.__init__(self, *args, **kwargs)
    
    def index(self):
        c.w.databaseView  = self.databaseView
        c.w.mainView=Widget("widget")
        return render_genshi('index.html', self.genshi_loader)

    
    def tableDef(self, id): #tableName):
        c.tableName = id
        sprocket  = self.sprockets['tableDef__'+c.tableName]
        c.w.mainView  = sprocket.view.widget
        c.w.databaseView  = self.databaseView
        c.mainValue = sprocket.session.getValue()
        for key, value in self.databaseDict.iteritems():
            setattr(c, key, value)
        return render_genshi('index.html', self.genshi_loader)
    
    def tableView(self, id):
        c.tableName = id
        c.page = request.params.get('page', 1)
        c.recordsPerPage = request.params.get('recordsPerPage', 25)
        #this should probably be a decorator
        c.page = int(c.page)
        c.recordsPerPage = int(c.recordsPerPage)
        sprocket  = self.sprockets['tableView__'+c.tableName]
        c.w.mainView  = sprocket.view.widget
        c.w.databaseView  = self.databaseView
        c.mainValue = sprocket.session.getValue(values=request.params, page=c.page, recordsPerPage=c.recordsPerPage)
        c.mainCount = sprocket.session.getCount(values=request.params)
        for key, value in self.databaseDict.iteritems():
            setattr(c, key, value)
        return render_genshi('tableView.html', self.genshi_loader)
    
    def addRecord(self, id): #tableName, **kw):
        c.tableName = id
        sprocket = self.sprockets['addRecord__'+c.tableName]
        c.w.mainView  = sprocket.view.widget
        c.w.databaseView  = self.databaseView
        kw = {}
        for key, value in request.params.iteritems():
            if key and value:
                setattr(kw, key, value)
        c.mainValue = sprocket.session.getValue(values=kw)
        for key, value in self.databaseDict.iteritems():
            setattr(c, key, value)
        return render_genshi('index.html', self.genshi_loader)
    
    def editRecord(self, id): #tableName, **kw):
        c.tableName = request.params.get('tableName', None)
        sprocket = self.sprockets['editRecord__'+c.tableName]
        c.w.mainView  = sprocket.view.widget
        c.w.databaseView  = self.databaseView
        kw = {}
        kw.update(request.params.__dict__)
        c.mainValue = sprocket.session.getValue(values=kw)
        for key, value in self.databaseDict.iteritems():
            setattr(c, key, value)
        return render_genshi('index.html', self.genshi_loader)
    
    def _createRelationships(self, tableName, params):
        #this might become a decorator
        #check to see if the table is a many-to-many table first
        if tableName in self.provider.getManyToManyTables():
            return
        #right now many-to-many only supports single primary keys
        id = params[self.provider.getPrimaryKeys(tableName)[0]]
        relationships = {}
        for key, value in params.iteritems():
            if key.startswith('many_many_'):
                relationships.setdefault(key[10:], []).append(value)
        for key, value in relationships.iteritems():
            self.provider.setManyToMany(tableName, id, key, value)
    
    # @validate(error_handler=editRecord)
    @crudErrorCatcher(errorType=sqlalchemy.exceptions.IntegrityError, error_handler='editRecord')
    @crudErrorCatcher(errorType=sqlalchemy.exceptions.ProgrammingError, error_handler='editRecord')
    def edit(self, id): # tableName, *args, **kw):
        tableName = request.params.get('tableName', None)
        params = pylons.request.params.copy()
        self._createRelationships(tableName, params)
        self.provider.edit(tableName, values=params)
        raise redirect(self.controller+'/tableView/'+tableName)
    
    # @validate(error_handler=addRecord)
    @crudErrorCatcher(errorType=sqlalchemy.exceptions.IntegrityError, error_handler='addRecord')
    @crudErrorCatcher(errorType=sqlalchemy.exceptions.ProgrammingError, error_handler='addRecord')
    def add(self, id): #tableName, **kw):
        tableName = request.params.get('tableName', None)
        params = pylons.request.params.copy()
        self._createRelationships(tableName, params)
        self.provider.add(tableName, values=params)
        raise redirect(self.controller+'/tableView/'+tableName)
    
    def delete(self, id): # tableName, **kw):
        tableName = request.params.get('tableName', None)
        self.provider.delete(tableName, values=kw)
        raise redirect(self.controller+'/tableView/'+tableName)



from nose.tools import raises

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from dbsprockets.metadata import Metadata, MetadataError, DatabaseMetadata, FieldsMetadata, FieldMetadata
from dbsprockets.saprovider import SAProvider
from dbsprockets.test.base import sortedTableList
from dbsprockets.iprovider import IProvider

import model
from model import *

def setup():
    engine = create_engine('sqlite://')
    metadata.bind = engine
    metadata.create_all(bind=engine)
    Session = sessionmaker(bind=engine, autoflush=True, transactional=True)
    session = Session()
    
    u = User()
    u.user_name=u'asdf'
    u.password = u'asdf'
    u.email=u"asdf@asdf.com"
    session.save(u)
    g = Group(group_name=u"group", display_name=u"group")
    session.save(g)
    session.flush()

provider = SAProvider(metadata)

class TestMetadata:
    obj = Metadata
    provider = IProvider()
    
    def setup(self):
        self.metadata = self.obj(self.provider)

    def testCreate(self):
        pass
    
    @raises(TypeError)
    def _create(self, arg1):
        Metadata(arg1)
        
    def testCreateBad(self):
        badInput = ((), [], {}, 'a', 1, 1.2)
        for input in badInput:
            yield self._create, input

    @raises(NotImplementedError)
    def testPrimaryKeys(self):
        self.metadata.primaryKeys()
    @raises(NotImplementedError)
    def testSetItem(self):
        self.metadata['asdf'] = 'asdf'
    @raises(NotImplementedError)
    def testGetItem(self):
        value = self.metadata['asdf']
    @raises(NotImplementedError)
    def testKeys(self):
        value = self.metadata.keys()
    @raises(NotImplementedError)
    def testAutoIncrementFields(self):
        value = self.metadata.autoIncrementFields

class TestDatabaseMetadata(TestMetadata):
    def setup(self):
        self.metadata = DatabaseMetadata(provider)
    def testKeys(self):
        tables = sorted(self.metadata.keys())
        expected = sortedTableList
        assert tables == expected, "expected: %s\n actual: %s"%(expected, tales)
    def testSetItem(self):
        self.metadata['asdf'] = 'asdf'
    def testGetItem(self):
        table = self.metadata['test_table']
    def testGetAddedInfo(self):
        self.metadata['asdf'] = 'asdf'
        assert self.metadata['asdf'] == 'asdf'
    @raises(MetadataError)
    def testSetItemBad(self):
        self.metadata['test_table'] = '1234'
    def testPrimaryKeys(self):
        assert [] == self.metadata.primaryKeys()
    def testGetForeignKeys(self):
        metadata = DatabaseMetadata(provider)
        pks = metadata.foreignKeys
        assert [] == pks, "%s"%pks
   
class TestFieldsMetadata(TestMetadata):
    def setup(self):
        self.metadata = FieldsMetadata(provider, 'test_table')
    def testGetItem(self):
        field = self.metadata['id']
        assert field.name == 'id'
    
    @raises(KeyError)
    def testGetItemKeyError(self):
        self.metadata['asdf']
        
    def testSetItem(self):
        self.metadata['asdf'] = '1234'
        assert self.metadata['asdf'] == '1234'
        
    @raises(MetadataError)
    def testSetItemBad(self):
        self.metadata['id'] = '1234'
    
    def testKeys(self):
        self.metadata['asdf'] = '1234'
        keys = sorted(self.metadata.keys())
        assert keys ==['BLOB', 'BOOLEAN_', 'Binary', 'Boolean', 'CHAR', 'CLOB', 'DATETIME_', 'DATE_', 'DECIMAL', 'Date', 'DateTime', 'FLOAT_', 'Float', 'INT', 'Integer', 'Numeric', 'Password', 'PickleType', 'SMALLINT', 'SmallInteger', 'String', 'TEXT', 'TIMESTAMP', 'TIME_', 'Time', 'Unicode', 'VARCHAR', 'asdf', 'created', 'id'], "actual: %s"%keys
        
    def testPrimaryKeys(self):
        pks = self.metadata.primaryKeys()
        assert ['id',] == pks, "%s"%pks

    def testGetForeignKeys(self):
        metadata = FieldsMetadata(provider, 'tg_user')
        pks = metadata.foreignKeys
        assert ['town',] == pks, "%s"%pks



"""Pylons application test package

When the test runner finds and executes tests within this directory,
this file will be loaded to setup the test environment.

It registers the root directory of the project in sys.path and
pkg_resources, in case the project hasn't been installed with
setuptools. It also initializes the application via websetup (paster
setup-app) with the project's test.ini configuration file.
"""
import os
import sys
from unittest import TestCase

import pkg_resources
import paste.fixture
import paste.script.appinstall
from paste.deploy import loadapp
from routes import url_for

__all__ = ['url_for', 'TestController']

here_dir = os.path.dirname(os.path.abspath(__file__))
conf_dir = os.path.dirname(os.path.dirname(here_dir))

sys.path.insert(0, conf_dir)
pkg_resources.working_set.add_entry(conf_dir)
pkg_resources.require('Paste')
pkg_resources.require('PasteScript')

test_file = os.path.join(conf_dir, 'test.ini')
cmd = paste.script.appinstall.SetupCommand('setup-app')
cmd.run([test_file])

class TestController(TestCase):

    def __init__(self, *args, **kwargs):
        wsgiapp = loadapp('config:test.ini', relative_to=conf_dir)
        self.app = paste.fixture.TestApp(wsgiapp)
        TestCase.__init__(self, *args, **kwargs)



from decorator import decorator
from tg import validate as tgValidate, flash, redirect
from pylons import request
from tg.controllers import _object_dispatch
class validate(tgValidate):
    def __init__(self,*args,**kwds):
        super(validate,self).__init__(*args,**kwds)
        class Validators(object):
            def validate(self, params):
                controller=request.environ["pylons.controller"]
                url_path = request.path.split('/')[1:]
                controller, remainder = _object_dispatch(controller, url_path)
                sprocket = controller.im_self.sprockets[params['dbsprockets_id']]
                res= sprocket.view.widget.validate(params)
                return res
        self.validators=Validators()

class validate_explicit(tgValidate):
    def __init__(self,sprockets,*args,**kwds):
        super(validate_explicit,self).__init__(*args,**kwds)
        class Validators(object):
            def validate(self, params):
                sprocket = sprockets[params['dbsprockets_id']]
                return sprocket.view.widget.validate(params)
        self.validators=Validators()
def crudErrorCatcher(errorType=None, error_handler=None):
    def wrapper(func, self, *args, **kwargs):
        """Decorator Wrapper function"""
        try:
            value = func(self, *args, **kwargs)
        except errorType, e:
            message=None
            if hasattr(e,"message"):
                message=e.message
            if isinstance(message,str):
                try:
                    message=message.decode('utf-8')
                except:
                    message=None
            if message:
                flash(message,status="status_alert")
            request.environ['pylons.routes_dict'] = error_handler
            return self._perform_call(None, dict(url=error_handler))
        return value
    return decorator(wrapper)



from nose.tools import eq_
from dbsprockets.widgets.widgets import *
from dbsprockets.test.base import *
from dbsprockets.saprovider import SAProvider

setupDatabase()
provider = SAProvider(metadata)

class TestContainerWidget:
    def setup(self):
        self.widget = ContainerWidget()
        
    def testCreateObj(self):
        pass
    
    def testDisplay(self):
        s = self.widget.render()
        assert 'class="containerwidget"' in s
        
class dummy(object):
    pass
class TestFieldViewWidget:
    def setup(self):
        d = dummy
        d.name = 'key'
        self.widget = RecordFieldWidget(identifier=d)
        
    def testCreateObj(self):
        pass
    
    def testDisplay(self):
        s = self.widget.render({'key':'value'})
        eq_(s, """<tr xmlns="http://www.w3.org/1999/xhtml" class="recordfieldwidget">
    <td>
        <b>key</b>
    </td>
    <td> value
    </td>
</tr>""")
        
class TestTableViewWidget:
    def setup(self):
        self.widget = TableLabelWidget()

    def testCreateObj(self):
        pass
    
    def testDisplay(self):
        s = self.widget.render()
        assert 'class="tablelabelwidget"' in s

class TestForeignKeySingleSelectField:
    def setup(self):
        self.widget = ForeignKeySingleSelectField(tableName='tg_user', provider=provider)

    def testCreateObj(self):
        pass
    
    def testDisplay(self):
        s = self.widget.render()
        assert s == """<select xmlns="http://www.w3.org/1999/xhtml" class="foreignkeysingleselectfield">
        <option value="1">asdf</option>
</select>""", s



try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

setup(
    name='TG2TestApp',
    version="",
    #description='',
    #author='',
    #author_email='',
    #url='',
    install_requires=[
        "TurboGears2",
        ],
    packages=find_packages(exclude=['ez_setup']),
    include_package_data=True,
    test_suite='nose.collector',
    package_data={'tg2testapp': ['i18n/*/LC_MESSAGES/*.mo',
                                 'templates/*/*',
                                 'public/*/*']},
    #message_extractors = {'tg2testapp': [
    #        ('**.py', 'python', None),
    #        ('templates/**.mako', 'mako', None),
    #        ('templates/**.html', 'genshi', None),
    #        ('public/**', 'ignore', None)]},
    entry_points="""
    [paste.app_factory]
    main = tg2testapp.config.middleware:make_app

    [paste.app_install]
    main = pylons.util:PylonsInstaller
    """,
)



from dbsprockets.util import MultiDict
from nose.tools import eq_


class testMultiDict:
    
    def setup(self):
        self.m = MultiDict()
        self.m['a'] = 1
        self.m['b'] = 2
        self.m['a'] = 3
        
    def testCreate(self):
        pass
    
    def testIterItems(self):
        actual = [(key, value) for key, value in self.m.iteritems()]
        eq_(actual, [('a', 1), ('a', 3), ('b', 2)])


from dbsprockets.view import View
from tw.api import Widget
from dbsprockets.viewconfig import ViewConfig
from dbsprockets.metadata import Metadata
from dbsprockets.iprovider import IProvider

class testView:
    def setup(self):
        provider = IProvider()
#        metadata = Metadata(provider)
        viewConfig = ViewConfig(provider, '')
        
        self.view = View(Widget(), viewConfig)
        
    def testCreate(self):
        pass
    
    def testDisplay(self):
        s = self.view.widget(value={})
        assert s == None, s


from nose.tools import raises
from dbsprockets.viewconfig import *
from dbsprockets.metadata import Metadata, DatabaseMetadata
from dbsprockets.saprovider import SAProvider
from dbsprockets.test.base import *
from dbsprockets.iprovider import IProvider
from tw.forms.fields import TextField
from formencode.schema import Schema

def setup():
    setupDatabase()

class DummyViewConfig(ViewConfig):
    pass
    
class DummyViewConfigWithRequiredFields(EditRecordViewConfig):
    _requiredFields = ['Integer',]

sqlmetadata = metadata
class TestViewConfig:
    obj = ViewConfig
    provider = IProvider()
    def setup(self):
        self.view = self.obj(self.provider, 'test_table')
        
    def testCreate(self):
        pass

    @raises(TypeError)
    def _create(self, arg1, arg2='', arg3=''):
        self.obj(arg1, arg2, arg3)
        
    def testCreateBad(self):
        provider = IProvider()
        badInput = ('a', (), {}, [], 1, 1.2)
        for input in badInput:
            yield self._create, input
        for input in badInput[1:]:
            yield self._create, provider, input
        for input in badInput[1:]:
            yield self._create, provider, '', input
            
    def testViewConfigWithLimitFieldsGetFields(self):
        self.view.limitFields = ['a',]
        assert self.view.fields == ['a',], self.view.fields

    @raises(NotImplementedError)
    def testGetWidgetArgs(self):
        self.view.getWidgetArgs()
        
    @raises(NotImplementedError)
    def testAction(self):
        self.view.action
        
    def testSetOmittedFields(self):
        self.view.omittedFields = []
    
    def testSetAdditionalFields(self):
        self.view.additionalFields = []

    def testSetDisabledFields(self):
        self.view.disabledFields = []

    def testSetRequiredFields(self):
        self.view.requiredFields = []

    def testSetHiddenFields(self):
        self.view.hiddenFields = []
        

class TestDatabaseViewConfig(TestViewConfig):
    provider = SAProvider(sqlmetadata)
    obj = DatabaseViewConfig
    def testGetWidgetArgs(self):
        self.view.getWidgetArgs()
    
class TestTableDefViewConfig(TestViewConfig):
    provider = SAProvider(sqlmetadata)
    obj = TableDefViewConfig

    def testGetWidgetArgs(self):
        self.view.getWidgetArgs()

class TestRecordViewConfig(TestViewConfig):
    provider = SAProvider(sqlmetadata)
    obj = RecordViewConfig

    def testGetWidgetArgs(self):
        args = self.view.getWidgetArgs()
        assert sorted(args.keys()) == ['children', 'controller', 'validator'], sorted(args.keys())

class TestTableViewConfig(TestViewConfig):
    provider = SAProvider(sqlmetadata)
    obj = TableViewConfig
    def testGetWidgetArgs(self):
        self.view.getWidgetArgs()
    
    def test_writePKsToURL(self):
        actual = self.view._writePKsToURL(['asdf','qwer'], {'asdf':'1234', 'qwer':'4567', 'ffff':'afd'}) 
        assert actual == '?asdf=1234&qwer=4567', actual
    
    def test_makeLinks(self):
        actual = self.view._makeLinks({'id':77}) 
        #print actual
        #i dont know why this test fails, but whatever
        #assert 'href' in actual, actual

class TestEditRecordViewConfig(TestViewConfig):
    provider = SAProvider(sqlmetadata)
    obj = EditRecordViewConfig
    def testGetWidgetArgs(self):
        self.view.getWidgetArgs()

    def testGetWidgetArgsViewConfigWithRequiredFieldsGetFields(self):
        viewConfig = DummyViewConfigWithRequiredFields(self.provider, 'test_table')
        args = viewConfig.getWidgetArgs()
        assert sorted(args.keys())==['action', 'children', 'controller', 'validator'], "%s"%args.keys()

    def testAction(self):
        assert self.view.action == '//edit', self.view.action
        
    def testFieldWidgets(self):
        self.view.fieldWidgets = {'Boolean':TextField}
        self.view.getWidgetArgs()

    def testFieldValidators(self):
        self.view.fieldValidators = {'Boolean':Schema}
        self.view.getWidgetArgs()

class TestAddRecordViewConfig(TestViewConfig):
    provider = SAProvider(sqlmetadata)
    obj = AddRecordViewConfig
    def testGetWidgetArgs(self):
        self.view.getWidgetArgs()

    def testAction(self):
        assert self.view.action == '//add', self.view.action

    def testDoGetManyToMany(self):
        view = self.obj(self.provider, identifier='tg_user')
        view.limitFields = ['tg_groups']
        view.getWidgetArgs()
        
    def testFieldAttrs(self):
        view = self.obj(self.provider, identifier='tg_user')
        view.fieldAttrs = {'display_name':{'rows':1}}
        attrs = view.getWidgetArgs()



from dbsprockets.test.base import setupDatabase, setupReflection, teardownDatabase
from dbsprockets.test.testSAProvider import _TestSAProvider

def setup():
    setupDatabase()
    setupReflection()
def teardown():
    teardownDatabase()
    
class TestSAProviderReflected(_TestSAProvider):
    pass



from dbsprockets.viewfactory import ViewFactory
from dbsprockets.viewconfig import ViewConfig
from dbsprockets.metadata import Metadata, NotFoundError
from dbsprockets.iprovider import IProvider

class DummyMetadata(Metadata):
    def _doKeys(self):
        return []
    

class DummyViewConfig(ViewConfig):
    metadataType = DummyMetadata
    
class testViewFactory:
    def setup(self):
        provider = IProvider()
        self.viewFactory = ViewFactory()
        self.viewConfig = DummyViewConfig(provider, 'test_table')
        
    def testCreateObj(self):
        pass
    
    def testCreate(self):
        view  = self.viewFactory.create(self.viewConfig)
    


import os
from sqlalchemy import *
from sqlalchemy.orm import *
from dbsprockets.test.model import *
from cStringIO import StringIO
from cgi import FieldStorage
sortedTableList = ['group_permission', 'permission', 'test_table', 'tg_group', 
                   'tg_user', 'town_table', 'user_group', 'visit', 'visit_identity', ]
import pkg_resources; 
pkg_resources.require('mysql-python')

session = None
engine = None
def setupDatabase():
    global session, engine
    engine = create_engine(os.environ.get('DBURL', 'sqlite://'))
    metadata.bind = engine
    metadata.drop_all()
    metadata.create_all()

    Session = sessionmaker(bind=engine, autoflush=True, transactional=True)
    session = Session()
    
    user = User()
    user.user_name = u'asdf'
    user.email = u"asdf@asdf.com"
    user.password = u"asdf"
    
    arvada = Town(name=u'Arvada')
    session.save(arvada)
    session.save(Town(name=u'Denver'))
    session.save(Town(name=u'Golden'))
    session.save(Town(name=u'Boulder'))
    session.commit()

    test_table.insert(values=dict(BLOB=FieldStorage('asdf', StringIO()).value)).execute()
    
    user.town = arvada.town_id
    session.save(user)
    
    for i in range (50):
        group = Group(group_name=unicode(i))
        session.save(group)
    
    user.groups.append(group)

    session.save(user)
    session.flush()
    session.commit()

def teardownDatabase():
    metadata.drop_all()

def _reassign_from_metadata():
    global visits_table, visit_identity_table, groups_table, town_table
    global users_table, permissions_table, user_group_table
    global group_permission_table, test_table

    visits_table = metadata.tables['visit']
    visit_identity_table = metadata.tables['visit_identity']
    groups_table = metadata.tables['tg_group']
    town_table = metadata.tables['town_table']
    users_table = metadata.tables['tg_user']
    permissions_table = metadata.tables['permission']
    user_group_table = metadata.tables['user_group']
    group_permission_table = metadata.tables['group_permission']
    test_table = metadata.tables['test_table']
    
def setupReflection():
    #if os.environ.get('AUTOLOAD', False):
    metadata.clear()
    metadata.reflect()
    _reassign_from_metadata()

    clear_mappers()
    tables = metadata.tables
    mapper(Town, tables['town_table'])
    mapper(Example, tables['test_table'])
    mapper(Visit, tables['visit'])
    mapper(VisitIdentity, tables['visit_identity'],
           properties=dict(users=relation(User, backref='visit_identity')))
    mapper(User, tables['tg_user'])
    mapper(Group, tables['tg_group'],
           properties=dict(users=relation(User,
                                          secondary=tables['user_group'],
                                          backref='groups')))
    mapper(Permission, tables['permission'],
           properties=dict(groups=relation(Group,
                                           secondary=tables['group_permission'],
                                           backref='permissions')))

    
if __name__ == '__main__':
    setupDatabase()



from nose.tools import raises
from dbsprockets.iprovider import IProvider

class TestIProvider:
    def setUp(self):
        self.provider = IProvider()
        
    def testCreate(self):
        pass
    
    @raises(NotImplementedError)
    def testGetTables(self):
        tables = sorted(self.provider.getTables())

    @raises(NotImplementedError)
    def testGetTable(self):
        table = self.provider.getTable('tg_user')

    @raises(NotImplementedError)
    def testGetColumns(self):
        columns = self.provider.getColumns('tg_user')

    @raises(NotImplementedError)
    def testGetColumn(self):
        column = self.provider.getColumn('tg_user', 'user_id')

    @raises(NotImplementedError)
    def testGetPrimaryKeys(self):
        keys = self.provider.getPrimaryKeys('tg_user')

    @raises(NotImplementedError)
    def testSelect(self):
        rows = self.provider.select('tg_user')
        
    @raises(NotImplementedError)
    def testAdd(self):
        self.provider.add()

    @raises(NotImplementedError)
    def testEdit(self):
        self.provider.edit()


"""
dbmechanic Module

this contains a turbogears controller which allows the user to have a 
phpMyAdmin *cringe*-like interface.  It is intended to be a replacement
for Catwalk

Classes:
Name                               Description
DBMechanic

Exceptions:
None

Functions:
None

Copywrite (c) 2007 Christopher Perkins
Original Version by Christopher Perkins 2007
Released under MIT license.
"""
import sqlalchemy
from tg.decorators import expose, validate
from tg.controllers import redirect
from tw.api import Widget, CSSLink
import pylons

from dbsprockets.sprockets import Sprockets
from tg import TGController
from dbsprockets.decorators import crudErrorCatcher, validate

dbsprocketsCss = CSSLink(modname='dbsprockets', filename='dbmechanic/static/css/dbmechanic.css')

class BaseController(TGController):
    """Basis TurboGears controller class which is derived from
    TGController
    """
    
    def __call__(self, environ, start_response):
        """Invoke the Controller"""
        # TGController.__call__ dispatches to the Controller method
        # the request is routed to. This routing information is
        # available in environ['pylons.routes_dict']

        # Create a container to send widgets to the template. Only those sent
        # in here will have their resources automatically included in the
        # template
        context.w = WidgetBunch()
        context.w.css = dbsprocketsCss
        try:
            return TGController.__call__(self, environ, start_response)
        finally:
            pass
            #after everything is done clear out the Database Session
            #so to eliminate possible cross request DBSession polution.
            #model.DBSession.remove()

class DBMechanic(BaseController):
    sprockets = None
    def __init__(self, provider, controller, *args, **kwargs):
        self.provider = provider
        self.sprockets = Sprockets(provider, controller)
        self.controller = controller
        
        #commonly used views
        sprocket = self.sprockets['databaseView']
        self.databaseValue = sprocket.session.getValue()
        self.databaseView  = sprocket.view.widget
        self.databaseDict  = dict(controller=self.controller)
        BaseController.__init__(self, *args, **kwargs)
        
    @expose('genshi:dbsprockets.dbmechanic.frameworks.tg2.templates.index')
    def index(self):
        pylons.c.w.databaseView  = self.databaseView
        pylons.c.w.mainView=Widget("widget")
        return self.databaseDict

    @expose('genshi:dbsprockets.dbmechanic.frameworks.tg2.templates.index')
    def tableDef(self, tableName):
        sprocket  = self.sprockets['tableDef__'+tableName]
        pylons.c.w.mainView  = sprocket.view.widget
        pylons.c.w.databaseView  = self.databaseView
        mainValue = sprocket.session.getValue()
        d = dict(tableName=tableName, mainValue=mainValue)
        d.update(self.databaseDict)
        return d
    
    @expose('genshi:dbsprockets.dbmechanic.frameworks.tg2.templates.tableView')
    def tableView(self, tableName, page=1, recordsPerPage=25, **kw):
        #this should probably be a decorator 
        page = int(page)
        recordsPerPage = int(recordsPerPage)
        
        sprocket  = self.sprockets['tableView__'+tableName]
        pylons.c.w.mainView  = sprocket.view.widget
        pylons.c.w.databaseView  = self.databaseView
        mainValue = sprocket.session.getValue(values=kw, page=page, recordsPerPage=recordsPerPage)
        mainCount = sprocket.session.getCount(values=kw,)
        d = dict(tableName=tableName, mainValue=mainValue, mainCount=mainCount)
        d.update(self.databaseDict)
        d['page'] = page
        d['recordsPerPage'] = recordsPerPage
        return d
    
    @expose('genshi:dbsprockets.dbmechanic.frameworks.tg2.templates.index')
    def addRecord(self, tableName, **kw):
        sprocket = self.sprockets['addRecord__'+tableName]
        pylons.c.w.mainView  = sprocket.view.widget
        pylons.c.w.databaseView  = self.databaseView
        mainValue = sprocket.session.getValue(values=kw)
        d = dict(tableName=tableName, mainValue=mainValue)
        d.update(self.databaseDict)
        return d
    
    @expose('genshi:dbsprockets.dbmechanic.frameworks.tg2.templates.index')
    def editRecord(self, tableName, **kw):
        sprocket = self.sprockets['editRecord__'+tableName]
        pylons.c.w.mainView  = sprocket.view.widget
        pylons.c.w.databaseView  = self.databaseView
        mainValue = sprocket.session.getValue(values=kw)
        d = dict(tableName=tableName, mainValue=mainValue)
        d.update(self.databaseDict)
        return d
    
    def createRelationships(self, tableName, params):
        #this might become a decorator
        #check to see if the table is a many-to-many table first
        if tableName in self.provider.getManyToManyTables():
            return
        #right now many-to-many only supports single primary keys
        pk=self.provider.getPrimaryKeys(tableName)
        assert len(pk)==1
        id = params[pk[0]]
        relationships = {}
        for key, value in params.iteritems():
            if key.startswith('many_many_'):
                relationships.setdefault(key[10:], []).append(value)
        for key, value in relationships.iteritems():
            self.provider.setManyToMany(tableName, id, key, value)
    
    @expose()
    @validate(error_handler=editRecord)
    @crudErrorCatcher(errorType=sqlalchemy.exceptions.IntegrityError, error_handler='editRecord')
    @crudErrorCatcher(errorType=sqlalchemy.exceptions.ProgrammingError, error_handler='editRecord')
    def edit(self, tableName, *args, **kw):
        params = pylons.request.params.copy()
        self.createRelationships(tableName, params)

        self.provider.edit(tableName, values=kw)
        redirect(self.controller+'/tableView/'+tableName)

    @expose()
    @validate(error_handler=addRecord)
    @crudErrorCatcher(errorType=sqlalchemy.exceptions.IntegrityError, error_handler='addRecord')
    @crudErrorCatcher(errorType=sqlalchemy.exceptions.ProgrammingError, error_handler='addRecord')
    def add(self, tableName, **kw):
        params = pylons.request.params.copy()
        self.createRelationships(tableName, params)

        self.provider.add(tableName, values=kw)
        redirect(self.controller+'/tableView/'+tableName)

    @expose()
    def delete(self, tableName, **kw):
        self.provider.delete(tableName, values=kw)
        redirect(self.controller+'/tableView/'+tableName)
    


"""Provides the BaseController class for subclassing, and other objects
utilized by Controllers.
"""
from pylons import c, cache, config, g, request, response, session
from tg import TGController, tmpl_context
from pylons.controllers.util import abort, etag_cache, redirect_to
from pylons.decorators import jsonify
from pylons.i18n import _, ungettext, N_
from pylons.templating import render
from tw.api import WidgetBunch

import tg2testapp.model as model

class Controller(object):
    """Base class for a web application's controller.
    
    Currently, this provides positional parameters functionality
    via a standard default method.
    """
    
class BaseController(TGController):
    """Base class for the root of a web application.
    
    Your web application should have one of these. The root of
    your application is used to compute URLs used by your app.
    """
    
    def __call__(self, environ, start_response):
        """Invoke the Controller"""
        # TGController.__call__ dispatches to the Controller method
        # the request is routed to. This routing information is
        # available in environ['pylons.routes_dict']

        # Create a container to send widgets to the template. Only those sent
        # in here will have their resources automatically included in the
        # template
        tmpl_context.w = WidgetBunch()
        try:
            return TGController.__call__(self, environ, start_response)
        finally:
            #after everything is done clear out the Database Session
            #so to eliminate possible cross request DBSession polution.
            model.DBSession.remove()



"""
dbmechanic Module

this contains a turbogears controller which allows the user to have a 
phpMyAdmin *cringe*-like interface.  It is intended to be a replacement
for Catwalk

Classes:
Name                               Description
DBMechanic

Exceptions:
None

Functions:
None

Copywrite (c) 2007 Christopher Perkins
Original Version by Christopher Perkins 2007
Released under MIT license.
"""
from tw.api import CSSLink, Link

#register all the resources in the static directory
dbMechanicCss = CSSLink(modname='dbsprockets', filename='dbmechanic/static/css/dbmechanic.css')
dbMechanicFooterImg = Link(modname='dbsprockets', filename='dbmechanic/static/images/tg_under_the_hood.png')


"""
validatorselecter Module

this contains the class which allows the ViewConfig to select the appropriate validator for the given field

Classes:
Name                               Description
ValidatorSelecter                     Parent Class
SAValidatorSelector                   Selecter Based on sqlalchemy field types
DatabaseViewValidatorSelector         Database View always selects the same validator
TableDefValidatorSelector             Table def fields use the same validator

Exceptions:
None

Functions:
None


Copywrite (c) 2007 Christopher Perkins
Original Version by Christopher Perkins 2007
Released under MIT license.
"""
from sqlalchemy.schema import Column
from sqlalchemy.types import *
from sqlalchemy.types import String as StringType
from tw.forms.validators import *
from formencode.compound import All
from formencode import Invalid

class UniqueValue(FancyValidator):
    def __init__(self, provider, field, *args, **kw):
        self.provider = provider
        self.field    = field
        FancyValidator.__init__(self, *args, **kw)

    def _to_python(self, value, state):
        if not self.provider.isUnique(self.field, value):
            raise Invalid(
                'That value already exists',
                value, state)
        return value

class ValidatorSelector(object):
    _nameBasedValidators = {}

    def __new__(cls, *args, **kw):
        bases = cls.mro()
        chainedAttrs = ['_nameBasedValidators']
        for base in bases:
            for attr in chainedAttrs:
                if hasattr(base, attr):
                    current =  getattr(cls, attr)
                    current.update(getattr(base, attr))
        return object.__new__(cls, *args, **kw)

    def __init__(self, *args, **kw):
        pass
    
    @property
    def nameBasedValidators(self):
        validators = self._doGetNameBasedValidators()
        validators.update(self._nameBasedValidators)
        return validators

    def select(self, field, required=False, checkIfUnique=False):
        return UnicodeString

    def _doGetNameBasedValidators(self):
        return {}
    
class SAValidatorSelector(ValidatorSelector):
    
    def __init__(self, provider):
        self.provider = provider
    
    defaultValidators = {
    StringType:   UnicodeString,
    Integer:  Int,
    Numeric:  Number,
    DateTime: DateValidator,
    Date:     DateValidator,
    Time:     DateValidator,
#    Binary:   UnicodeString,
    PickleType: UnicodeString,
#    Boolean: UnicodeString,
#    NullType: TextField
    }
    _nameBasedValidators = {'email_address':Email}

    def select(self, field, required=False, checkIfUnique=False):
        if not isinstance(field, Column):
            raise TypeError("arg1 must be a sqlalchemy column, not %s"%type(field))
        
        #do not validate boolean or binary arguments
        if isinstance(field.type, (Boolean, Binary)):
            return None
        
        validatorArgs = {}
        validatorArgs['not_empty']=False
        if required:
            validatorArgs['not_empty']=True
        

        if field.name in self.nameBasedValidators:
            return self.nameBasedValidators[field.name](**validatorArgs)

        tipe = StringType
        for t in self.defaultValidators.keys():
            if isinstance(field.type, t):
                tipe = t
                break
        
        validatorType = self.defaultValidators[tipe]
        if hasattr(field.type,'length') and validatorType is UnicodeString:
            validatorArgs['max']=field.type.length
        validator = validatorType(**validatorArgs)
            
        if field.unique and checkIfUnique:
            validator = All(UniqueValue(self.provider, field), validator)
        return validator


"""
SessionConfig Module

The SessionConfig defines all of the Data related issues for dbsprockets

Classes:
Name                               Description
SessionConfig                      Parent Class
DatabaseSessionConfig
TableViewSessionConfig
AddRecordViewConfig
EditRecordSessionConfig

Exceptions:
None

Functions:
None

Copywrite (c) 2007 Christopher Perkins
Original Version by Christopher Perkins 2007
Released under MIT license.
"""
from cStringIO import StringIO
import types
from iprovider import IProvider
from dbsprockets.metadata import Metadata, DatabaseMetadata, FieldsMetadata, FieldMetadata
from dbsprockets.util import MultiDict, Label

class SessionConfig(object):
    metadataType = Metadata
    def __init__(self, id, provider, identifier=None):
        if not isinstance(id, types.StringTypes):
            raise TypeError('id is not a string')
        if not isinstance(provider, IProvider):
            raise TypeError('provider is not of type IProvider')
        if identifier is not None and not isinstance(identifier, types.StringTypes):
            raise TypeError('identifier is not a string')
        self.id = id
        self.provider = provider
        self.identifier = identifier
        self.metadata = self.metadataType(provider, identifier)

    def getValue(self, values={}):
        """get the value associated with the session
        Arguments:
        values : as provided from the browser
        """
        values = self._doGetValue(values=values)
        return values

    def getCount(self, values={}):
        return self._doGetCount(values)
    
    def _doGetCount(self, values={}):
        raise NotImplementedError
    
    def _doGetValue(self, values={}):
        #the default getValue is just a pass through
        return values

class DatabaseSessionConfig(SessionConfig):
    metadataType = DatabaseMetadata
    
    def _doGetValue(self, values={}):
        r = {}
        tables = self.metadata.keys()
        for table in tables:
            r[table] = self.metadata[table]
        return r

class TableViewSessionConfig(SessionConfig):
    metadataType = FieldsMetadata
    foreignKeyFieldLabels = ['name', '_name', 'description', '_description', 'title']
    
    def _doGetForeignKeyDict(self):
        return self.provider.getForeignKeyDict(self.identifier, self.foreignKeyFieldLabels)
    
    #for this class it makes a lot of sense to override the parent method
    def getValue(self, values={}, page=1, recordsPerPage=20):   
        manyToManyTables = self.provider.getAssociationTables(self.identifier)
        tableName = self.identifier
        offset = (page - 1)*recordsPerPage
        rows = self.provider.select(tableName, resultOffset=offset, resultLimit=recordsPerPage)
        #this is probably going to have to be changed (too slow)
        foreignKeyDict = self._doGetForeignKeyDict()
        primaryKeys=self.provider.getPrimaryKeys(self.identifier)
        newRows = rows
        if len(rows) > 0:
            newRows = []
            for row in rows:
                d = {}
                for key in row.keys():
                    value = row[key]
                    d[key] = value
                    if key == 'password':
                        d[key] = '*'*6
                    if value is not None and key in foreignKeyDict and value in foreignKeyDict[key]:# and not key in primaryKeys:
                        d[key] = Label(foreignKeyDict[key][value])
                        d[key].original=value
                    if self.provider.isBinaryColumn(self.identifier, key):
                        d[key] = '<file>'
                        if row[key] is None:
                            d[key] = ''

                if self.identifier not in manyToManyTables:
                    manyToManyColumns = self.provider.getManyToManyColumns(self.identifier)
                    for column in manyToManyColumns:
                        tableName = column[:-1] #strip off the 's'
                        manyToManyTable = self.provider.getManyToManyTable(self.identifier, tableName)
                        sourcePK = self.provider.getPrimaryKeys(self.identifier)[0]
                        values = {sourcePK:row[sourcePK]}
                        viewColumn = self.provider.getViewColumnName(tableName)
                        idColumn   = self.provider.getPrimaryKeys(tableName)[0]
                        selectedThings = self.provider.select(manyToManyTable, values=values, columnsLimit=[idColumn,])
                        values = MultiDict()
                        for item in selectedThings:
                            values[idColumn] = item[idColumn]
                        manyToManyRows = []
                        if len(selectedThings) != 0:
                            manyToManyRows = self.provider.select(tableName, values=values, whereclauseJoin='or', columnsLimit=[viewColumn,])
                        manyToManyLabels = ''
                        for value in manyToManyRows:
                            manyToManyLabels += unicode(value[viewColumn]) +', '
                        manyToManyLabels = manyToManyLabels[:-2]
                        d[column] = manyToManyLabels
                newRows.append(d)
            
        return newRows
    
    def _doGetCount(self, values={}):
        return self.provider.count(self.identifier)
        
    
class AddRecordSessionConfig(SessionConfig):
    metadataType = FieldsMetadata
    
    def _doGetValue(self, values={}):
        #attach the tablename to the  values
        values['tableName'] = self.metadata.identifier
        values['dbsprockets_id'] =  self.id
        return values
    
class EditRecordSessionConfig(AddRecordSessionConfig):
    metadataType = FieldsMetadata
    
    def _doGetManyToMany(self, values={}):
        #most of this should probably go in provider
        tableName = self.identifier
        pk = self.provider.getPrimaryKeys(tableName)[0]
        srcTable = self.provider.getTable(tableName)

        associationTables = self.provider.getAssociationTables(tableName)
        for table in associationTables:
            table = self.provider.getTable(table)
            srcColumn, dstColumn = table.c
            if srcColumn.foreign_keys[0].column.table != srcTable:
                temp = srcColumn
                srcColumn = dstColumn
                dstColumn = temp
                table = srcColumn.table
            d = {srcColumn.name:values[str(pk)]}
            rows = self.provider.select(table.name, values=d)
            newValues = [row[dstColumn] for row in rows]
            values[str('many_many_'+dstColumn.foreign_keys[0].column.table.name)] = newValues
        return values
    
    def _doGetValue(self, values={}):
        if self.identifier not in self.provider.getManyToManyTables():
            values.update(self._doGetManyToMany(values))
        values = super(EditRecordSessionConfig, self)._doGetValue(values)
        #sql object is not attachable, but a dictionary is

        row = self.provider.selectOnPks(tableName=self.identifier, values=values)[0]
        values.update(row)
        return values


"""
widgetselecter Module

this contains the class which allows the ViewConfig to select the appropriate widget for the given field

Classes:
Name                               Description
WidgetSelecter                     Parent Class
SAWidgetSelector                   Selecter Based on sqlalchemy field types
DatabaseViewWidgetSelector         Database View always selects the same widget
TableDefWidgetSelector             Table def fields use the same widget

Exceptions:
None

Functions:
None


Copywrite (c) 2007 Christopher Perkins
Original Version by Christopher Perkins 2007
Released under MIT license.
"""
from sqlalchemy.schema import Column
from sqlalchemy.types import *

from tw.api import Widget
from tw.forms.fields import *

from dbsprockets.widgets.widgets import *
from dbsprockets.saprovider import SAProvider

class WidgetSelector:
    def select(self, field):
        return Widget
    
class DatabaseViewWidgetSelector(WidgetSelector):
    def select(self, field):
        return TableLabelWidget

class TableDefWidgetSelector(WidgetSelector):
    def select(self, field):
        return TableDefWidget

    
class RecordViewWidgetSelector(WidgetSelector):
    def select(self, field):
        return RecordFieldWidget

text_field_limit=100
        
class SAWidgetSelector(WidgetSelector):
    
    defaultWidgets = {
    String:   TextField,
    Integer:  TextField,
    Numeric:  TextField,
    DateTime: DBSprocketsCalendarDateTimePicker,
    Date:     DBSprocketsCalendarDatePicker,
    Time:     DBSprocketsTimePicker,
    Binary:   FileField,
    PickleType: TextField,
    Boolean: DBSprocketsCheckBox,
#    NullType: TextField
    }
    
    defaultNameBasedWidgets = {}
    
    def _getSelectWidget(self, field):
        return ForeignKeySingleSelectField
    
    def select(self, field):
        if not isinstance(field, Column):
            raise TypeError("arg1 must be a sqlalchemy column, not %s"%type(field))
        
        if isinstance(field.type, Integer):
            if len(field.foreign_keys) == 1:
                return self._getSelectWidget(field)

        if field.name in self.defaultNameBasedWidgets:
            return self.defaultNameBasedWidgets[field.name]
            
        if field.name.lower() == 'password':
            return PasswordField

        tipe = String
        for t in self.defaultWidgets.keys():
            if isinstance(field.type, t):
                tipe = t
                break

        widget = self.defaultWidgets[tipe]
        if widget is TextField and hasattr(field.type, 'length') and (field.type.length is None or field.type.length>text_field_limit):
            widget = TextArea
        return widget
    


from pylons import config
from datetime import datetime
from sqlalchemy import *#Column, MetaData, Table, types
from sqlalchemy.orm import mapper, relation
from sqlalchemy.orm import scoped_session, sessionmaker

# Global session manager.  Session() returns the session object
# appropriate for the current web request.
DBSession = scoped_session(sessionmaker(autoflush=True, transactional=True))

# Global metadata. If you have multiple databases with overlapping table
# names, you'll need a metadata for each database.
metadata = MetaData()

def init_model(d):
    pass

# The identity schema.
visits_table = Table('visit', metadata,
    Column('visit_key', String(40), primary_key=True),
    Column('created', DateTime, nullable=False, default=datetime.now),
    Column('expiry', DateTime)
)

visit_identity_table = Table('visit_identity', metadata,
    Column('visit_key', String(40), primary_key=True),
    Column('user_id', Integer, ForeignKey('tg_user.user_id'), index=True)
)

groups_table = Table('tg_group', metadata,
    Column('group_id', Integer, primary_key=True),
    Column('group_name', Unicode(16), unique=True),
    Column('display_name', Unicode(255)),
    Column('created', DateTime, default=datetime.now)
)

town_table = Table('town_table', metadata,
    Column('town_id', Integer, primary_key=True),
    Column('name', Unicode(255)))

users_table = Table('tg_user', metadata,
    Column('user_id', Integer, primary_key=True),
    Column('user_name', Unicode(16), unique=True),
    Column('email_address', Unicode(255), unique=True),
    Column('display_name', Unicode(255)),
    Column('password', Unicode(40)),
    Column('town', Integer, ForeignKey('town_table.town_id')),
    Column('created', DateTime, default=datetime.now)
)
permissions_table = Table('permission', metadata,
    Column('permission_id', Integer, primary_key=True),
    Column('permission_name', Unicode(16), unique=True),
    Column('description', Unicode(255))
)

user_group_table = Table('user_group', metadata,
    Column('user_id', Integer, ForeignKey('tg_user.user_id'), primary_key=True),
    Column('group_id', Integer, ForeignKey('tg_group.group_id'), primary_key=True)
)

group_permission_table = Table('group_permission', metadata,
    Column('group_id', Integer, ForeignKey('tg_group.group_id'), primary_key=True),
    Column('permission_id', Integer, ForeignKey('permission.permission_id'), primary_key=True)
)

test_table = Table('test_table', metadata,
                   Column('id', Integer, primary_key=True),
                   Column('created', DateTime, default=datetime.now),
                   Column('BLOB',         BLOB          ),
                   Column('BOOLEAN_',      BOOLEAN       ),
                   Column('Binary',       Binary        ),
                   Column('Boolean',      Boolean       ),
                   Column('CHAR',         CHAR(200)     ),
                   Column('CLOB',         CLOB(200)     ),
                   Column('DATE_',         DATE          ),
                   Column('DATETIME_',     DATETIME      ),
                   Column('DECIMAL',      DECIMAL       ),
                   Column('Date',         Date          ),
                   Column('DateTime',     DateTime      ),
                   Column('FLOAT_',        FLOAT         ),
                   Column('Float',        Float         ),
                   Column('INT',          INT           ),
                   Column('Integer',      Integer       ),
        #           Column('NCHAR',        NCHAR         ),
                   Column('Numeric',      Numeric       ),
                   Column('PickleType',   PickleType    ),
                   Column('SMALLINT',     SMALLINT      ),
                   Column('SmallInteger', SmallInteger  ),
                   Column('String',       String        ),
                   Column('TEXT',         TEXT          ),
                   Column('TIME_',         TIME          ),
                   Column('Time',         Time          ),
                   Column('TIMESTAMP',    TIMESTAMP     ),
                   Column('Unicode',      Unicode       ),
                   Column('VARCHAR',      VARCHAR(200)  ),
                   )

class Example(object):
    pass

class Visit(object):
    def lookup_visit(cls, visit_key):
        return Visit.get(visit_key)
    lookup_visit = classmethod(lookup_visit)

class VisitIdentity(object):
    pass

class Group(object):
    """
    An ultra-simple group definition.
    """
    pass

class User(object):
    """
    Reasonably basic User definition. Probably would want additional
    attributes.
    """
    def __init__(self, user_name=None,
                 password=None,
                 email=None):
        self.user_name = user_name
        self.password = password
        self.email = email

    def permissions(self):
        perms = set()
        for g in self.groups:
            perms = perms | set(g.permissions)
        return perms
    permissions = property(permissions)

class Permission(object):
    pass

mapper( Example, test_table)
mapper( Visit, visits_table)
mapper( VisitIdentity, visit_identity_table,
          properties=dict(users=relation(User, backref='visit_identity')))
mapper( User, users_table)
mapper( Group, groups_table,
          properties=dict(users=relation(User,secondary=user_group_table, backref='groups')))
mapper( Permission, permissions_table,
          properties=dict(groups=relation(Group,secondary=group_permission_table, backref='permissions')))



"""Pylons environment configuration"""
import os

from pylons import config

from pylons.i18n import ugettext
from genshi.filters import Translator
from tg import defaults
from sqlalchemy import engine_from_config

import tg2testapp.lib.app_globals as app_globals
from tg2testapp.model import init_model

def template_loaded(template):
    "Plug-in our i18n function to Genshi."
    template.filters.insert(0, Translator(ugettext))

def load_environment(global_conf, app_conf):
    """Configure the Pylons environment via the ``pylons.config``
    object
    """
    # Pylons paths
    root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    paths = dict(root=root,
                 controllers=os.path.join(root, 'controllers'),
                 static_files=os.path.join(root, 'public'),
                 templates=[os.path.join(root, 'templates')])

    # This setups up a set of default route that enables a standard
    # TG2 style object dispatch.   Fell free to overide it with
    # custom routes.  TODO: Link to TG2+routes doc.

    make_map = defaults.make_default_route_map

    # Initialize config with the basic options
    config.init_app(global_conf, app_conf, package='tg2testapp',
                    template_engine='genshi', paths=paths)
    config['routes.map'] = make_map()
    config['pylons.app_globals'] = app_globals.Globals()
    config['pylons.app_globals'].sa_engine = engine_from_config(config, 'sqlalchemy.')

    # If you'd like to change the default template engine used to render
    # text/html content, edit these options.
    template_engine = 'genshi'
    template_engine_options = {}

    config['buffet.template_engines'].pop()
    config.add_template_engine(template_engine, 'tg2testapp.templates',
                               template_engine_options)

    # Pass the engine to initmodel, to be able to introspect tables
    engine = engine_from_config(config, 'sqlalchemy.')
    init_model(engine)

    # CONFIGURATION OPTIONS HERE (note: all config options will override
    # any Pylons config options)

    from tg2testapp import model
    model.DBSession.configure(bind=config['pylons.app_globals'].sa_engine)
    model.metadata.bind = config['pylons.app_globals'].sa_engine




import pylons
from tg2testapp.lib.base import BaseController
from tg import expose, flash
from pylons.i18n import ugettext as _
from tg import redirect, validate
from tw.forms.fields import PasswordField
from formencode.validators import FieldsMatch
from formencode import Schema

from tg2testapp.model import User, metadata
from dbsprockets.dbmechanic.frameworks.tg2 import DBMechanic
from dbsprockets.saprovider import SAProvider
from dbsprockets.primitives import makeForm, makeTable, getTableValue

requiredFields = ['password', 'user_name', 'email_address']
limitFields  = ['user_name', 'display_name', 'email_address', 'password', ]

additionalFields = [PasswordField('passwordVerification', label_text='Verify'),]
formValidator =  Schema(chained_validators=(FieldsMatch('password',
                                                        'passwordVerification',
                                                        messages={'invalidNoMatch': 
                                                                  "Passwords do not match"}),))

registrationForm = makeForm(User, 
                            'registration', 
                            requiredFields=requiredFields, 
                            limitFields=limitFields, 
                            additionalFields=additionalFields, 
                            formValidator=formValidator)
usersTable       = makeTable(User, '/', omittedFields=['user_id', 'created', 'password'])
loginForm        = makeForm (User, identifier='myLoginForm', action='/login', limitFields=['user_name', 'password'])

class RootController(BaseController):

    dbmechanic = DBMechanic(SAProvider(metadata), '/dbmechanic')

    @expose('sprockets.templates.index')
    def index(self):
        from datetime import datetime
        flash(_("Your application is now running"))
        return dict(now=datetime.now())

    @expose('genshi:sprockets.templates.register')
    def register(self, **kw):
        pylons.c.w.form = registrationForm
        return dict(value=kw)

    @validate(registrationForm, error_handler=register)
    def registration(self, **kw):
        raise redirect('/')

    
    @expose('genshi:sprockets.templates.register')
    def users(self):
        pylons.c.w.form = usersTable
        value = getTableValue(User)
        return dict(value=value)
    
    @expose('genshi:sprockets.templates.register')
    def login(self, **kw):
        pylons.c.w.form = loginForm
        return dict(value=kw)



"""The application's Globals object"""
from pylons import config

class Globals(object):
    """Globals acts as a container for objects available throughout the
    life of the application
    """

    def __init__(self):
        """One instance of Globals is created during application
        initialization and is available during requests via the 'g'
        variable
        """
        pass



"""
dbmechanic Module (tg1 compat)

this contains a turbogears controller which allows the user to have a 
phpMyAdmin *cringe*-like interface.  It is intended to be a replacement
for Catwalk

Classes:
Name                               Description
DBMechanic

Exceptions:
None

Functions:
None

Copywrite (c) 2007 Christopher Perkins
Original Version by Christopher Perkins 2007
Released under MIT license.
"""
import sqlalchemy
import cherrypy
from tw.api import Widget, CSSLink, Link


try:
    from turbogears.controllers import Controller
    from turbogears import expose, redirect, validate, error_handler
    import turbogears
    #register the cherrypy static directory for the image and css files.
    def add_custom_stdvars(vars):
        return vars.update({"dbmech_css": dbMechanicCss,
                            "dbmech_footer_img": dbMechanicFooterImg,
                           })
    
    turbogears.view.variable_providers.append(add_custom_stdvars)
except ImportError:
    # when nosetests is going through package discovery, it hits this module.
    # since there is not turbogears application running in this python instance
    # it crashes saying that turbgears.config is not defined.
    # we do this so that nosetests can load a dummy version of this module and keep on trucking.
    class Controller: pass
    def expose(*args, **kw):
        return lambda *args, **kw: None
    def error_handler(*args, **kw):
        return lambda *args, **kw: None
    def validate(*args, **kw):
        return lambda *args, **kw: None

from tw.api import Widget
from dbsprockets.sprockets import Sprockets
from dbsprockets.saprovider import SAProvider
#from dbsprockets.validator import validate, crudErrorCatcher

from dbsprockets.dbmechanic.resources import dbMechanicCss, dbMechanicFooterImg



def getSprocket(self):
    request = cherrypy.request
    sprocket = self.sprockets[request.params['dbsprockets_id']]
    form = sprocket.view.widget
    return form

class DBMechanic(Controller):
    def __init__(self, provider, controller, *args, **kwargs):
        self.provider = provider
        self.sprockets = Sprockets(provider, controller)
        self.controller = controller
        
        #commonly used views
        sprocket = self.sprockets['databaseView']
        self.databaseValue = sprocket.session.getValue()
        self.databaseView  = sprocket.view.widget
        self.databaseDict  = dict(databaseView=self.databaseView, databaseValue=self.databaseValue, controller=self.controller)#, databaseView=self.databaseView, databaseValue=self.databaseValue)
        

    @expose('genshi:dbsprockets.dbmechanic.frameworks.tg.templates.index')
    def index(self):
        d = dict(mainView=Widget("widget"), mainValue = {}, tableName='')
        d.update(self.databaseDict)
        return d


    @expose('genshi:dbsprockets.dbmechanic.frameworks.tg.templates.index')
    def tableDef(self, tableName):
        sprocket  = self.sprockets['tableDef__'+tableName]
        mainValue = sprocket.session.getValue()
        d = dict(tableName=tableName, mainView=sprocket.view.widget, mainValue=mainValue)
        d.update(self.databaseDict)
        return d
    
    @expose('genshi:dbsprockets.dbmechanic.frameworks.tg.templates.tableView')
    def tableView(self, tableName, page=1, recordsPerPage=25, **kw):
        #this should probably be a decorator 
        page = int(page)
        recordsPerPage = int(recordsPerPage)
        
        sprocket  = self.sprockets['tableView__'+tableName]
        mainValue = sprocket.session.getValue(values=kw, page=page, recordsPerPage=recordsPerPage)
        mainCount = sprocket.session.getCount(values=kw)
        d = dict(tableName=tableName, mainView=sprocket.view.widget, mainValue=mainValue, mainCount=mainCount)
        d.update(self.databaseDict)
        d['page'] = page
        d['recordsPerPage'] = recordsPerPage
        return d
    
    @expose('genshi:dbsprockets.dbmechanic.frameworks.tg.templates.index')
    def addRecord(self, tableName, **kw):
        sprocket = self.sprockets['addRecord__'+tableName]
        mainValue = sprocket.session.getValue(values=kw)
        d = dict(tableName=tableName, mainView=sprocket.view.widget, mainValue=mainValue)
        d.update(self.databaseDict)
        return d
    
    @expose('genshi:dbsprockets.dbmechanic.frameworks.tg.templates.index')
    def editRecord(self, tableName, **kw):
        sprocket = self.sprockets['editRecord__'+tableName]
        mainValue = sprocket.session.getValue(values=kw)
        d = dict(tableName=tableName, mainView=sprocket.view.widget, mainValue=mainValue)
        d.update(self.databaseDict)
        return d
    
    def createRelationships(self, tableName, params):
        #this might become a decorator
        #check to see if the table is a many-to-many table first
        if tableName in self.provider.getManyToManyTables():
            return
        #right now many-to-many only supports single primary keys
        params = cherrypy.request.params
        id = params[self.provider.getPrimaryKeys(tableName)[0]]
        relationships = {}
        for key, value in params.iteritems():
            if key.startswith('many_many_'):
                relationships.setdefault(key[10:], []).append(value)
        for key, value in relationships.iteritems():
            self.provider.setManyToMany(tableName, id, key, value)
    
    @expose()
    @validate(getSprocket)
#    @crudErrorCatcher(errorType=sqlalchemy.exceptions.IntegrityError, error_handler='editRecord')
    @error_handler(editRecord)
    def edit(self, tableName, *args, **kw):
        params = cherrypy.request
        self.createRelationships(tableName, params)

        self.provider.edit(tableName, values=kw)
        raise redirect(self.controller+'/tableView/'+tableName)

    @expose()
    @validate(getSprocket)
    @error_handler(addRecord)
#    @crudErrorCatcher(errorType=sqlalchemy.exceptions.IntegrityError, error_handler='addRecord')
    def add(self, tableName, **kw):
        params = cherrypy.request
        self.createRelationships(tableName, params)

        self.provider.add(tableName, values=kw)
        raise redirect(self.controller+'/tableView/'+tableName)

    @expose()
    def delete(self, tableName, **kw):
        self.provider.delete(tableName, values=kw)
        raise redirect(self.controller+'/tableView/'+tableName)
    


"""
matadata Module

this contains the class which definies the generic interface for metadata
that is used in form generation later in the pipeline

Classes:
Name                               Description
Metadata                           general metadata function
DatabaseMetadata                   describes the fields as entries in the dictionary
FieldsMetadata                     describes the columns as entries in the dictionary
FieldMetadata                      generic type this is not used right now.

Exceptions:
MetadataError

Functions:
None

Copywrite (c) 2007 Christopher Perkins
Original Version by Christopher Perkins 2007
Released under MIT license.
"""
from dbsprockets.iprovider import IProvider

class MetadataError(Exception):pass
class NotFoundError(Exception):pass

class Metadata(dict):
    """
    """
    def __init__(self, provider, identifier=None):
        if not isinstance(provider, IProvider):
            raise TypeError('provider must be of type IProvider not %s'%type(provider))
        self.provider = provider
        self.identifier=identifier
        
    def __setitem__(self, key, value):
        self._doCheckSetItem(key, value)
        dict.__setitem__(self, key, value)

    def __getitem__(self, item):
        try:
            value = self._doGetItem(item)
            return value
        except NotFoundError:
            return dict.__getitem__(self, item)

    def keys(self):
        r = self._doKeys()
        r.extend(dict.keys(self))
        return r
    def primaryKeys(self):
        return self._doGetPrimaryKeys()
    
    @property
    def foreignKeys(self):
        return self._doGetForeignKeys()
    
    @property
    def autoIncrementFields(self):
        return self._doGetAutoIncrementFields()
    
    def _doGetAutoIncrementFields(self):
        raise NotImplementedError
    def _doGetForeignKeys(self):
        raise NotImplementedError
    def _doGetPrimaryKeys(self):
        raise NotImplementedError
    def _doCheckSetItem(self, key, value):
        raise NotImplementedError
    def _doGetItem(self, item):
        raise NotImplementedError
    def _doKeys(self):
        raise NotImplementedError

class DatabaseMetadata(Metadata):
    """
    """
    def __init__(self, provider, identifier=None):
        Metadata.__init__(self, provider, identifier)
        self._tables = None
        
    def _doCheckSetItem(self, key, value):
        if key in self.provider.tables:
            raise MetadataError('%s is already a table name in the database'%key)
        
    def _doGetPrimaryKeys(self):
        return []
    
    def _doGetForeignKeys(self):
        return []

    def _doGetItem(self, item):
        if item in self.provider.tables:
            return self.provider.getTable(item)
        raise NotFoundError

    def _doKeys(self):
        if self._tables is None:
            self._tables = self.provider.getTables()
        return self._tables
    
class FieldsMetadata(Metadata):
    """
    """
    def __init__(self, provider, identifier):
        Metadata.__init__(self, provider, identifier)
        self.table = self.provider.getTable(identifier)

    def _doCheckSetItem(self, key, value):
        if key in self.table.columns:
            raise MetadataError('%s is already found in table: %s'%(key, self.table))

    def _doGetItem(self, item):
        if item in self.table.columns:
            return self.table.columns[item]
        raise NotFoundError

    def _doGetAutoIncrementFields(self):
        return self.provider.getAutoIncrementFields(self.identifier)
    
    def _doGetPrimaryKeys(self):
        return self.provider.getPrimaryKeys(self.identifier)
    
    def _doGetForeignKeys(self):
        return [c.name for c in self.provider.getForeignKeys(self.identifier)]
    
    def _doKeys(self):
        r = self.table.columns.keys()
        return r
        
class FieldMetadata(Metadata):
    """
    """
    pass




import os, sys, os.path
import pkg_resources

from routes import url_for
from sqlalchemy.orm import sessionmaker, scoped_session
from tg.tests.base import TestWSGIController, make_app
from webob import Request, Response
from paste.deploy import loadapp
from paste.fixture import TestApp
from pprint import pprint
from nose.tools import eq_

app = None

def setup():
    global app
    here_dir = os.path.dirname(os.path.abspath(__file__))
    proj_dir = os.path.join(here_dir, 'TG2TestApp')
    
    pkg_resources.working_set.add_entry(proj_dir)

    from tg2testapp.model import DBSession, metadata, User, Group
    app = loadapp('config:development.ini', relative_to=proj_dir)
    app = TestApp(app)
    metadata.drop_all()
    metadata.create_all()

    session = DBSession
    user = User()
    user.user_name = u'asdf'
    user.email = u"asdf@asdf.com"
    user.password = u"asdf"
    
    session.save(user)
    
    for i in range (50):
        group = Group()
        group.group_name=unicode(i)
        session.save(group)
        user.groups.append(group)

    session.save(user)
    session.commit()
    session.flush()
    
def teardown():
    from tg2testapp.model import User, Group, DBSession as session, metadata
    del metadata

class TestDBMechanic:
    def setup(self):
        from tg2testapp.model import users_table, user_group_table, groups_table
        users_table.delete().execute()
        groups_table.delete().execute()
        user_group_table.delete().execute()
        
    def testDatabaseView(self):
        resp = app.get('/dbmechanic')
        s =  resp.body
        assert """<div class="tablelabelwidget">
<a href="/dbmechanic/tableView/tg_user">tg_user</a>
</div>
<div class="tablelabelwidget">
<a href="/dbmechanic/tableView/town_table">town_table</a>
</div>
<input type="hidden" name="dbsprockets_id" class="hiddenfield" id="databaseView_dbsprockets_id" value="">
</div>""" in s, "actual: %s"%s
    
    def testTableDef(self):
        resp = app.get('/dbmechanic/tableDef/tg_user')
        value =  resp.body
        assert """<tr class="tabledefwidget">
    <td>
        created
    </td>
    <td>
    DateTime(timezone=False)
    </td>
</tr>
<input type="hidden" name="dbsprockets_id" class="hiddenfield" id="tableDef__tg_user_dbsprockets_id" value="">
</table>
      </div>""" in value, value
        
    def testTableView(self):
        from tg2testapp.model import User, Group, DBSession as session
        u = User()
        u.user_name=u'userTableView'
        u.password = u'asdf'
        u.email_address=u"asdftableView@asdf.com"
        session.save(u)
        group = Group()
        group.group_name=u"table_view_group"
        group.display_name=u"group"
        session.save(group)
        session.flush()
        
        resp = app.get('/dbmechanic/tableView/tg_user')
        value =  resp.body
        assert """<thead>
        <tr>
            <th class="col_0">
            </th><th class="col_1">
            user_id
            </th><th class="col_2">
            user_name
            </th><th class="col_3">
            email_address
            </th><th class="col_4">
            display_name
            </th><th class="col_5">
            password
            </th><th class="col_6">
            town
            </th><th class="col_7">
            created
            </th><th class="col_8">
            tg_groups
            </th>
        </tr>
    </thead>
    <tbody>
        <tr class="even">
            <td><a href="/dbmechanic/editRecord/tg_user?user_id=1">"""
        """edit</a>|<a href="/dbmechanic/delete/tg_user?user_id=1">delete</a></td><td>1</td><td>user2</td>"""
        """<td></td><td></td><td>******</td><td>""" in value, value
        
    def testAddRecord(self):
        resp = app.get('/dbmechanic/addRecord/tg_user')
        value =  resp.body
        assert """<td>
                <div>
    <input type="text" id="addRecord__tg_user_created" class="dbsprocketscalendardatetimepicker" name="created" value="2007-12-21 19:02:25" />
    <input type="button" id="addRecord__tg_user_created_trigger" class="date_field_button" value="Choose" />
</div>
            </td>
        </tr><tr class="even">
            <th>
            </th>
            <td>
                <input type="submit" class="submitbutton" id="addRecord__tg_user_submit" value="Submit" />
            </td>
        </tr>
    </table>"""
        
    def testEditRecord(self):
        from tg2testapp.model import User, Group, DBSession as session
        u = User()
        u.user_name=u'user3'
        u.password = u'asdf'
        u.email_address=u"asdf@asdf2.com"
        session.save(u)
        group = Group()
        group.group_name=u"table_view_group"
        group.display_name=u"group"
        session.save(group)
        session.flush()
        resp = app.get('/dbmechanic/editRecord/tg_user?user_id=1')
        value =  resp.body
        assert """<td>
                <div>
    <input type="text" id="editRecord__tg_user_created" class="dbsprocketscalendardatetimepicker" name="created" value="2007-12-21 20:45:00" />
    <input type="button" id="editRecord__tg_user_created_trigger" class="date_field_button" value="Choose" />
</div>
            </td>
        </tr><tr class="even">
            <th>
            </th>
            <td>
                <input type="submit" class="submitbutton" id="editRecord__tg_user_submit" value="Submit" />
            </td>
        </tr>
    </table>
</form>"""

#cant test this until we can do it properly through paste
    def testAdd(self):
        resp = app.get('/dbmechanic/addRecord/tg_user')
        value =  resp.body
        resp = app.get('/dbmechanic/add/tg_user', params=dict(user_id=None,
                                                              user_name=u'new_user',
                                                              email_address=u'new@user.com',
                                                              dbsprockets_id=u'addRecord__tg_user'))
        #actual =  resp.follow().namespace
        actual = resp.namespace
#        eq_({}, value)
        
        expected = {'controller': '/dbmechanic', 
                    #'context': <paste.registry.StackedObjectProxy object at 0x2225f90>, 
                    #'recordsPerPage': 25, 
                    'tableName': 'tg_user', 
                    #'mainCount': 1, 
                    #'page': 1
                    }
        mainValue ={u'town': None, 
                    u'user_id': 1, 
#                           u'created': datetime.datetime(2008, 2, 23, 11, 59, 35, 144897), 
                    u'email_address': u'new@user.com', 
                    'tg_groups': '', 
                    u'display_name': None, 
                    u'password': '******', 
                    u'user_name': u'new_user'}
        
        for key, value in expected.iteritems():
            yield eq_, value, actual[key]
        
        #actualRow = actual['mainValue'][0]

#        for key, value in mainValue.iteritems():
#            yield eq_, value, actualRow[key]

    def testEdit(self):
        from tg2testapp.model import User, Group, DBSession as session
        u = User()
        u.user_name=u'userTestEdit'
        u.password = u'asdf'
        u.email_address=u"test@edit.com"
        session.save(u)
        session.commit()
        session.flush()
        u_id = str(u.user_id)
        resp = app.get('/dbmechanic/editRecord/tg_user?user_id='+u_id)
        value =  resp.body
        resp = app.get('/dbmechanic/edit/tg_user', params=dict(user_id=u_id, 
                                                              user_name=u'new_username',
                                                              email_address=u'newemail@user.com',
                                                              dbsprockets_id=u'editRecord__tg_user'
                                                          ))
        actual =  resp.follow().namespace
        expected = {'controller': '/dbmechanic', 
                    'recordsPerPage': 25, 
                    'tableName': 'tg_user', 
                    'mainCount': 1, 
                    'page': 1}
        mainValue ={u'town': None, 
                    u'user_id': 1, 
#                           u'created': datetime.datetime(2008, 2, 23, 11, 59, 35, 144897), 
                    u'email_address': u'newemail@user.com', 
                    'tg_groups': '', 
                    u'display_name': None, 
                    u'password': '******', 
                    u'user_name': u'new_username'}
        
        for key, value in expected.iteritems():
            yield eq_, value, actual[key]
        actualRow = actual['mainValue'][0]

        for key, value in mainValue.iteritems():
            yield eq_, value, actualRow[key]
            
    def testCreateRelationships(self):
        from tg2testapp.model import User, Group, DBSession as session
        u = User()
        u.user_name=u'userCreateRel'
        u.password = u'asdf'
        u.email_address=u"create@relationships.com"
        session.save(u)
        session.commit()
        group = Group()
        group.group_name=u"create_rel_group"
        group.display_name=u"group"
        session.save(group)
        session.flush()
        u_id = str(u.user_id)
        #add group 1
        resp = app.get('/dbmechanic/edit/tg_user', params=dict(user_id=u_id, 
                                                        user_name=u'new_username',
                                                        email_address=u'newemail@user.com',
                                                        dbsprockets_id=u'editRecord__tg_user',
                                                        many_many_tg_group="1",
                                                          ))

    def testCreateRelationshipsManyMany(self):
        from tg2testapp.model import user_group_table
        user_group_table.insert(values=dict(user_id=1, group_id=2)).execute()
        resp = app.get('/dbmechanic/edit/user_group', params=dict(dbsprockets_id='editRecord__user_group', user_id=1, group_id=2))

    def testEdit(self):
        from tg2testapp.model import User, Group, DBSession as session
        u = User()
        u.user_name=u'userTestEdit'
        u.password = u'asdf'
        u.email_address=u"test@edit.com"
        session.save(u)
        session.commit()
        session.flush()
        u_id = str(u.user_id)
        resp = app.get('/dbmechanic/delete/tg_user?user_id='+u_id)
        
        



from tw.api import Widget
from tw.forms import CalendarDatePicker, CalendarDateTimePicker, TableForm
from tw.forms.fields import SingleSelectField, MultipleSelectField, InputField
from formencode.schema import Schema
from formencode.validators import Bool

class DBSprocketsCalendarDatePicker(CalendarDatePicker):
    date_format = '%Y-%m-%d'

class DBSprocketsTimePicker(CalendarDateTimePicker):
    date_format = '%H:%M:%S'

class DBSprocketsCalendarDateTimePicker(CalendarDateTimePicker):
    date_format = '%Y-%m-%d %H:%M:%S'

class ContainerWidget(Widget):
    template = "genshi:dbsprockets.widgets.templates.container"
    params = ["controller",]

class TableLabelWidget(Widget):
    template = "genshi:dbsprockets.widgets.templates.tableLabel"
    params = ["identifier", "controller"]
    
class RecordViewWidget(Widget):
    template = "genshi:dbsprockets.widgets.templates.recordViewTable"
    params = ["identifier"]

class RecordFieldWidget(Widget):
    template = "genshi:dbsprockets.widgets.templates.recordField"
    params = ['identifier', 'controller']

class TableDefWidget(Widget):
    template = "genshi:dbsprockets.widgets.templates.tableDef"
    params = ["identifier"]

class TableWidget(Widget):
    template = "genshi:dbsprockets.widgets.templates.table"
    
class DBSprocketsTableForm(TableForm):
    validator = Schema(ignore_missing_keys=True, allow_extra_fields=True)
    template = "genshi:dbsprockets.widgets.templates.tableForm"
    
    def update_params(self, d):
        super(DBSprocketsTableForm, self).update_params(d)
        
#custom checkbox widget since I am not happy with the behavior of the TW one
class DBSprocketsCheckBox(InputField):
    template = "genshi:dbsprockets.widgets.templates.checkbox"
    validator = Bool
    def update_params(self, d):
        InputField.update_params(self, d)
        try:
            checked = self.validator.to_python(d.value)
        except Invalid:
            checked = False
        d.attrs['checked'] = checked or None

class ForeignKeyMixin(Widget):
    params = ["tableName", "provider"]
    def _my_update_params(self, d,nullable=False):
        viewColumn = self.provider.getViewColumnName(self.tableName)
        idColumn = self.provider.getIDColumnName(self.tableName)
        rows = self.provider.select(self.tableName, columnsLimit=[idColumn, viewColumn])
        rows= [(row[idColumn], row[viewColumn]) for row in rows]
        if nullable:
            rows.append([None,"-----------"])
        if len(rows) == 0:
            return {}
        d['options']= rows
        
        return d
    
class ForeignKeySingleSelectField(SingleSelectField, ForeignKeyMixin):
    params=["nullable"]
    nullable=False
    def update_params(self, d):
        self._my_update_params(d,nullable=self.nullable)
        SingleSelectField.update_params(self, d)
        return d

class ForeignKeyMultipleSelectField(MultipleSelectField, ForeignKeyMixin):
    def update_params(self, d):
        self._my_update_params(d)
        MultipleSelectField.update_params(self, d)
        return d



"""
saprovider Module

this contains the class which allows dbsprockets to interface with sqlalchemy.

Classes:
Name                               Description
SAProvider                         sqlalchemy metadata/crud provider

Exceptions:
None

Functions:
None


Copywrite (c) 2007 Christopher Perkins
Original Version by Christopher Perkins 2007
Released under MIT license.
"""

from sqlalchemy import *
from dbsprockets.iprovider import IProvider
from cgi import FieldStorage


class SAProvider(IProvider):
    """A Class to manipulate an sqlalchemy interface.
    
    Functions:
    getTables      get a list of tables from the database
    getTable       get the table definition for a given table
    getColumns     get the columns from a given table
    getPrimaryKeys get the primary keys from a given table.
    
    Attributes:
    metadata       link to the sqlalchemy metadata
    """
    whereclauseJoinFuncs = {'and':and_, 'or':or_}
    
    def __init__(self, metadata):
        if not isinstance(metadata, MetaData):
            raise TypeError('metadata is not of type sqlalchemy.MetaData')
        self.metadata = metadata
        self.tables = {}
        for name, table in metadata.tables.iteritems():
            self.tables[name] = table
        
    def getTables(self):
        """()->[list of tablesNames] 
        get a list of tables from the database"""
        return self.metadata.tables.keys()
    tables = property(getTables)
    
    def getTable(self, name):
        """(name) -> sqlalchemy.schema.Table
        get the table definition with the given table name
        """
        return self.metadata.tables[name]
    
    def getColumns(self, tableName):
        """(name) -> [list of columnNames]
        get the column names given the table name
        """
        return self.metadata.tables[tableName].columns.keys()
    
    def getColumn(self, tableName, name):
        """(tableName, name) -> sqlalchemy.schema.Column
        get the column definition givcen the tablename and column name.
        """
        return self.metadata.tables[tableName].columns[name]
    
    def getPrimaryKeys(self, tableName):
        """(name) -> [list of columnNames that are primary keys]
        get a list of columns which are primary keys
        """
        table = self.getTable(tableName)
        return table.primary_key.keys()

    def getAutoIncrementFields(self, tableName):
        table = self.getTable(tableName)
        return [c.name for c in table.c if c.autoincrement]
    
    def getDefaultValues(self, tableName):
        table = self.getTable(tableName)
        r = {}
        for c in table.c:
            if c.default is not None:
                r[c.name] = c.default.execute()
        return r
    
    def _generateWhereclause(self, table, pks, values={}, join='and'):
        l = [(getattr(table.c, key)==value) for key, value in values.iteritems() if key in pks and hasattr(table.c, key)]
        whereclause = None
        #one primary key
        if len(l) == 1:
            whereclause = l[0]
        #case of multiple keys
        elif len(l) > 1:
            joinFunc = self.whereclauseJoinFuncs[join]
            whereclause = joinFunc(*l)
        return whereclause
    
    def _generateColumnListing(self, table, columnNames):
        l = [getattr(table.c, key) for key in columnNames if hasattr(table.c, key)]
        return l

    def _removePrimaryKeys(self, tableName, values={}):
        for key in self.getPrimaryKeys(tableName):
            if key in values:
                values.pop(key)
        return values
    
    def _removePrimaryKeysIfNotForeign(self, tableName, values={}):
        foreignKeys=[c.name for c in self.getForeignKeys(tableName)]
        for key in self.getPrimaryKeys(tableName):
            if (key in values) and (not (key in foreignKeys)):
                values.pop(key)
        return values
    
    
    def _removeNonLimitColumns(self, values, limit):
        if limit is None:
            return values
        for key in values.keys():
            if key not in limit:
                values.pop(key)
        return values

    def getForeignKeys(self, tableName):
        table = self.getTable(tableName)
        return [column for column in table.columns if len(column.foreign_keys)>0]

    def getForeignKeyDict(self, tableName, foreignKeyFieldLabels=['name', '_name', 'description', '_description', 'title']):
        """Returns a dict of dicts where the keys are the column names, and then the row values.  Something like this:
        {town_table:{1:'Arvada', 2:'Denver', 3:'Golden'}}
        """
        d = {}
        foreignKeyColumns = self.getForeignKeys(tableName)
        for column in foreignKeyColumns:
            table = column.foreign_keys[0].column.table
            foreignTableName = table.name
            rows      = table.select().execute().fetchall()
            if len(rows) == 0:
                d[column.name] = {}
                continue
            viewColumn = self._findFirstColumn(foreignTableName, foreignKeyFieldLabels)
            idColumn   = self._findFirstColumn(foreignTableName, ['_id', 'id'])
            innerD = {}
            for row in rows:
                innerD[row[idColumn]] = row[viewColumn]
            d[column.name] = innerD
        return d
    
    def getManyToManyTables(self):
        if not hasattr(self, '_manyToManyTables'):
            self._manyToManyTables = [table for table in self.tables if len(self.getForeignKeys(table)) == 2 and len(self.getColumns(table)) == 2]
        return self._manyToManyTables

    def getManyToManyColumns(self, tableName):
        table = self.getTable(tableName)
        manyToManyTables = self.getAssociationTables(tableName)
        manyToManyColumns = []
        for manyToManyTable in manyToManyTables:
            column1, column2 = self.getTable(manyToManyTable).columns
            if column1.foreign_keys[0].column.table == table:
                manyToManyColumns.append(column2.foreign_keys[0].column.table.name+'s')
            else:
                manyToManyColumns.append(column1.foreign_keys[0].column.table.name+'s')
        return manyToManyColumns
    
    def getManyToManyTable(self, tableName1, tableName2):
        table1 = self.getTable(tableName1)
        table2 = self.getTable(tableName2)
        for table in self.getManyToManyTables():
            tables = [column.foreign_keys[0].column.table for column in self.getTable(table).columns]
            if table1 in tables and table2 in tables:
                return table
            
    
    def getAssociationTables(self, tableName):
        tables = []
        srcTable = self.getTable(tableName)
        for table in self.getManyToManyTables():
            for column in self.getTable(table).columns:
                key = column.foreign_keys[0]
                if key.column.table is srcTable:
                    tables.append(table)
                    break
        return tables
    
    def setManyToMany(self, srcTableName, srcID, relatedTableName, values):
        srcTable = self.getTable(srcTableName)
        relatedTable = self.getTable(relatedTableName)
        
        manyToManyTables = self.getManyToManyTables()
        foundTable = None
        #find the right table to modify
        for table in manyToManyTables:
            table = self.getTable(table)
            column1, column2 = table.c
            if column1.foreign_keys[0].column.table == srcTable and column2.foreign_keys[0].column.table == relatedTable:
                foundTable = table
                break;
            if column2.foreign_keys[0].column.table == srcTable and column1.foreign_keys[0].column.table == relatedTable:
                foundTable = table
                (column1,column2)=(column2,column1)
                break;
        if foundTable is None:
            return
        #clear it out
        foundTable.delete(whereclause=column1==srcID).execute()
        
        #add the new values
        for value in values:
            foundTable.insert(values={column1:srcID, column2:value}).execute()
        
        #this operation was successful
        return True
        
    def getAssociatedManyToManyTables(self, tableName):
        tables =self.getAssociationTables(tableName)
        srcTable = self.getTable(tableName)
        r = []
        for table in tables:
            for column in self.getTable(table).columns:
                key = column.foreign_keys[0]
                if key.column.table is not srcTable:
                    r.append(key.column.table.name)
                    continue
        return r
    
    def _findFirstColumn(self, tableName, possibleColumns):
        actualColumns = self.getColumns(tableName)
        viewColumn = None
        for columnName in possibleColumns:
            for actualName in actualColumns:
                if columnName in actualName:
                    viewColumn = actualName
                    break
            if viewColumn:
                break;
        if viewColumn is None:
            viewColumn = actualColumns[0]
        return viewColumn
    
    def isBinaryColumn(self, tableName, columnName):
        return columnName in self.getColumns(tableName) and isinstance(self.getColumn(tableName, columnName).type, Binary)
    
    def getViewColumnName(self, tableName, possibleColumns=['_name', 'name', 'description', 'title']):
        return self._findFirstColumn(tableName, possibleColumns)

    def getIDColumnName(self, tableName):
        possibleColumns = ['_id', 'id']
        return self._findFirstColumn(tableName, possibleColumns)

    def _select(self, tableName, pks, values, columnsLimit, resultLimit, resultOffset, whereclauseJoin='and'):
        order_by = pks
        table = self.getTable(tableName)
        whereclause = self._generateWhereclause(table, pks, values, whereclauseJoin)
        columnListing=[table]
        if columnsLimit is not None:
            columnListing = self._generateColumnListing(table, columnsLimit)
            order_by      = columnListing
        s = select(columnListing, whereclause).offset(resultOffset).limit(resultLimit).order_by(*list(table.c))
        return s.execute().fetchall()
        
    def isUnique(self, field, value):
        s = select([field,], whereclause=field==value).execute().fetchall()
        return len(s) == 0

    #crud below
    def selectOnPks(self, tableName, values={}, columnsLimit=None, resultLimit=None, resultOffset=None):
        pks = self.getPrimaryKeys(tableName)
        return self._select(tableName, pks, values, columnsLimit, resultLimit, resultOffset)

    def select(self, tableName, values={}, columnsLimit=None, resultLimit=None, resultOffset=None, whereclauseJoin='and'):
        pks = values.keys()
        return self._select(tableName, pks, values, columnsLimit, resultLimit, resultOffset, whereclauseJoin)
    
    def count(self, tableName, values=None):
        table = self.getTable(tableName)
        return table.select().alias('all_data').count().execute().fetchall()[0][0]
         
    def add(self, tableName=None, values={}, columnsLimit=None):
        for key, value in values.iteritems():
            if isinstance(value, FieldStorage):
                values[key] = value.value
                
        #remove the primary keys which could cause a conflict
        kw = self._removePrimaryKeysIfNotForeign(tableName, values)
        #kw=self._removePrimaryKeys(tableName, values)
        from sys import stderr
        values = self._removeNonLimitColumns(values, columnsLimit)
        table = self.getTable(tableName)
        table.insert(values=kw).execute()
        
    def delete(self, tableName, values={}):
        table = self.getTable(tableName)
        pks = self.getPrimaryKeys(tableName)
        whereclause = self._generateWhereclause(table, pks, values)
        return table.delete(whereclause).execute()
        
    def edit(self, tableName, values={}, columnsLimit=None):
        pks = self.getPrimaryKeys(tableName)
        table = self.getTable(tableName)
        values = self._removeNonLimitColumns(values, columnsLimit)
        whereclause = self._generateWhereclause(table, pks, values)
        #remove the primary keys which could cause a conflict
        kw = self._removePrimaryKeys(tableName, values)
        table = self.getTable(tableName)
        if len(kw)>0:
            table.update(values=kw, whereclause=whereclause).execute()
    def isNullableField(self,field):
        return field.nullable


"""Setup the TG2TestApp application"""
import logging

from paste.deploy import appconfig
from pylons import config

from tg2testapp.config.environment import load_environment

log = logging.getLogger(__name__)

def setup_config(command, filename, section, vars):
    """Place any commands to setup tg2testapp here"""
    conf = appconfig('config:' + filename)
    load_environment(conf.global_conf, conf.local_conf)

    # Load the models
    from tg2testapp import model
    print "Creating tables"
    model.metadata.create_all(bind=config['pylons.app_globals'].sa_engine)
    model.DBSession.commit()
    print "Successfully setup"



"""
primitives Module

short cut functions for dbsprocket form creation

Classes:
Name                               Description
None

Exceptions:
None

Functions:
makeForm        make a form for user entry
makeTable       make a table for user values
makeView        view helper function
getTableValues  helper to get table values


Copywrite (c) 2007 Christopher Perkins
Original Version by Christopher Perkins 2007
Released under MIT license.
"""
import new

from dbsprockets.saprovider import SAProvider
from dbsprockets.viewconfig import RecordViewConfig, TableViewConfig, EditableRecordViewConfig
from dbsprockets.viewfactory import ViewFactory
from dbsprockets.sessionconfig import TableViewSessionConfig
import types
from formencode.schema import Schema
from tw.forms import TableForm
import sqlalchemy
from dbsprockets.util import freshdefaults

class DBHelper:
    def __init__(self):
        self._identifiers = {}
        self._metadata = None
        self._provider = None

    def getMetadata(self, model):
        raise NotImplementedError
    
    def getIdentifier(self, model):
        raise NotImplementedError

    def validateModel(self, model):
        raise NotImplementedError
    
class _SAORMDBHelper(DBHelper):
        
    def _firstColumn(self, model):
        return model.c[model.c.keys()[0]]
    
    def getMetadata(self, model):
        firstColumn = self._firstColumn(model)
        if self._metadata is None:
            self._metadata = firstColumn.metadata
            self._identifiers[model] = firstColumn.table.name
        return self._metadata
    
    def getIdentifier(self, model):
        if model not in self._identifiers:
            self._identifiers[model] = self._firstColumn(model).table.name
        return self._identifiers[model]

    def getProvider(self, model):
        if self._provider is None:
            self._provider = SAProvider(self.getMetadata(model))
        return self._provider
    
    def validateModel(self, model):
        if not hasattr(model, 'c'):
            raise TypeError('arg1(model) has not been mapped to an sql table appropriately')
        
        #failsafe, return a form with nothing in it
        if len(model.c.keys()) == 0:
            raise Exception('Form has no fields')

SAORMDBHelper = _SAORMDBHelper()

#XXX:
#StormHelper = _StormHelper()
#SOHelper    = _SOHelper()

class DatabaseMixin:
    def __init__(self, dbHelper=None):
        self._dbHelper = dbHelper
        
    def _getHelper(self, model):
        if self._dbHelper is None:
            #use a SA Helper
            if hasattr(model, 'c') and isinstance(model.c, sqlalchemy.util.OrderedProperties):
                self._dbHelper = SAORMDBHelper
            #other helper definitions are going in here
            else:
                raise Exception('arg1(model) has not been mapped to an sql table appropriately')
        return self._dbHelper
    
    def _validateModel(self, model):
        helper = self._getHelper(model)
        helper.validateModel(model)

class SprocketGenerator(DatabaseMixin):
    
    def makeForm(self, 
                 model, 
                 action, 
                 identifier='',
                 controller='/',
                 hiddenFields=None,
                 disabledFields=None,
                 requiredFields=None, 
                 omittedFields=None,
                 additionalFields=None,
                 limitFields=None,
                 fieldAttrs=None,
                 widgets=None, 
                 validators=None, 
                 formWidgetType=None,
                 formValidator=None,
                 checkIfUnique=False,
                 ):
        """
        makeForm(model,
                 action,
                 identifier='',
                 identifier='',
                 controller='/',
                 hiddenFields=[],
                 disabledFields=[],
                 requiredFields=[], 
                 omittedFields=[],
                 additionalFields=[],
                 limitFields=None,
                 fieldAttrs={},
                 widgets={},
                 validators={}, 
                 formWidgetType=None,
                 formValidator=None
                 checkIfUnique=False)-->tw.forms.TableForm
                 
        Generates an entry form for use in web applications.
        
        Arguments:
        name              type                           default    Description
        model             class                                     sqlalchemy mapped class definition
        action            String                                    string passed to the forms 'action'
        identifier        String or None                 None       optional identifier for the form
        hiddenFields      list if names                  []         fields that are hidden but not altogether 
                                                                    removed from the table, useful for passing 
                                                                    data through a form manually.
        disabledFields    list of names                  []         fields that are shown but not editable
        requiredFields    list of names                  []         fields that cannot be left empty by the user (validation type stuff)
        omittedFields     list of names                  []         fields that are not present, whatsoever
        additionalFields  list of Toscawidgets.fields    []         fields that you would like to add to the form that are not in the model
        limitFields       list of names                  None       limit the fields to the one in the list.  A side effect is that the 
                                                                    order of the fields is set by this argument.
                                                                    (a good way to order your fields if you include all of them)
        fieldAttrs        dict of field Attrs            {}         a set of extra parameters which can be passed into the widgets
                                                                    that are created in a form.
        widgets           dict of Toscawidgets           {}         a set of widgets linked to your fields by fieldname, you must do your own validation.
        validators        dict of formencode.validators  {}         validators linked to fields in your form indexed by fieldname
        formWidgetType    ToscaWidget                    None       field for the actual widget to be used
        formValidator     Schema or validator            None       validator for the overall form
        checkIfUnique     Boolean                        False      adds default validator to "unique" fields to check if the value
                                                                    the user has entered is unique or not.
        """
        if hiddenFields is None:
            hiddenFields = []
        if disabledFields is None:
            disabledFields=[]
        if requiredFields is None:
            requiredFields=[]
        if omittedFields is None:
            omittedFields=[]
        if additionalFields is None:
            additionalFields=[]
        if fieldAttrs is None:
            fieldAttrs={}
        if widgets is None:
            widgets={}
        if validators is None:
            validators={}
        
        
        if not isinstance(action, types.StringTypes):
            raise TypeError('arg2(action) must be of type String, not %s'%type(action))
        if not isinstance(controller, types.StringTypes):
            raise TypeError('arg3(controller) must be of type String, not %s'%type(action))
        if not isinstance(identifier, types.StringTypes):
            raise TypeError('arg4(action) must be of type String, not %s'%type(action))
        if not isinstance(hiddenFields, types.ListType):
            raise TypeError('arg5(hiddenFields) must be of type List, not %s'%type(hiddenFields))
        if not isinstance(disabledFields, types.ListType):
            raise TypeError('arg6(disabledFields) must be of type List, not %s'%type(disabledFields))
        if not isinstance(omittedFields, types.ListType):
            raise TypeError('arg7(omittedFields) must be of type List, not %s'%type(omittedFields))
        if not isinstance(additionalFields, types.ListType):
            raise TypeError('arg8(additionalFields) must be of type List, not %s'%type(additionalFields))
        if limitFields is not None and not isinstance(limitFields, types.ListType):
            raise TypeError('arg9(limitFields) must be of type List, not %s'%type(limitFields))
        if not isinstance(fieldAttrs, types.DictType):
            raise TypeError('arg10(fieldAttrs) must be of type Dict, not %s'%type(limitFields))
        if not isinstance(validators, types.DictType):
            raise TypeError('arg11(validators) must be of type Dict, not %s'%type(validators))
        if formValidator is not None and not isinstance(formValidator, Schema):
            raise TypeError('arg12(formValidator) must be of type Dict, not %s'%type(formValidator))
        if not isinstance(checkIfUnique, bool):
            raise TypeError('arg13(checkIfUnique) must be of type bool, not %s'%type(checkIfUnique))
        
        self._validateModel(model)
        
        view = self.makeView(EditableRecordViewConfig, 
                        model, 
                        action,
                        identifier,
                        '',
                        hiddenFields, 
                        disabledFields,
                        requiredFields, 
                        omittedFields, 
                        additionalFields,
                        limitFields,
                        fieldAttrs,
                        widgets,
                        validators,
                        formWidgetType,
                        formValidator,
                        checkIfUnique)
        return view.widget
    
    def makeRecordView(self, 
                 model, 
                 controller='',
                 omittedFields=None,
                 additionalFields=None,
                 limitFields=None,
                 fieldAttrs=None,
                 widgets=None, 
                 validators=None, 
                 formWidgetType=None,
                 ):
        """
        makeRecordView(model, 
                 controller='',
                 omittedFields=None,
                 additionalFields=None,
                 limitFields=None,
                 fieldAttrs=None,
                 widgets=None, 
                 validators=None, 
                 formWidgetType=None,
                )-->Toscawidgets.TableForm
                 
        Generates an entry form for use in web applications.
        
        Arguments:
        name              type                           default    Description
        model             class                                     sqlalchemy mapped class definition
        identifier        String or None                 None       optional identifier for the form
        controller        String or None                 None       helps with doing links, possibly
        omittedFields     list of names                  []         fields that are not present, whatsoever
        additionalFields  list of Toscawidgets.fields    []         fields that you would like to add to the form that are not in the model
        limitFields       list of names                  None       limit the fields to the one in the list.  A side effect is that the 
                                                                    order of the fields is set by this argument.
                                                                    (a good way to order your fields if you include all of them)
        fieldAttrs        dict of field Attrs            {}         a set of extra parameters which can be passed into the widgets
                                                                    that are created in a form.
        formWidgetType    ToscaWidget                    None       field for the actual widget to be used
        """
        if omittedFields is None:
            omittedFields=[]
        if additionalFields is None:
            additionalFields=[]
        if fieldAttrs is None:
            fieldAttrs={}
        if widgets is None:
            widgets={}
        if validators is None:
            validators={}
        
        if not isinstance(omittedFields, types.ListType):
            raise TypeError('arg7(omittedFields) must be of type List, not %s'%type(omittedFields))
        if not isinstance(additionalFields, types.ListType):
            raise TypeError('arg8(additionalFields) must be of type List, not %s'%type(additionalFields))
        if limitFields is not None and not isinstance(limitFields, types.ListType):
            raise TypeError('arg9(limitFields) must be of type List, not %s'%type(limitFields))
        if not isinstance(fieldAttrs, types.DictType):
            raise TypeError('arg10(fieldAttrs) must be of type Dict, not %s'%type(limitFields))
        
        self._validateModel(model)
        
        view = self.makeView(RecordViewConfig, 
                        model, 
                        '',
                        '',
                        controller,
                        omittedFields=omittedFields,
                        additionalFields=additionalFields,
                        limitFields=limitFields,
                        fieldAttrs=fieldAttrs,
                        widgets=widgets,
                        widgetType=formWidgetType,
                         )
        return view.widget

    def makeView(self, viewConfigType, 
                 model, 
                 action='', 
                 identifier=None,
                 controller='/',
                 hiddenFields=None, 
                 disabledFields=None,
                 requiredFields=None,
                 omittedFields=None, 
                 additionalFields=None,
                 limitFields=None,
                 fieldAttrs=None,
                 widgets=None, 
                 validators=None,
                 widgetType=None,
                 formValidator=None,
                 checkIfUnique=False,
                 makeLinks=None
                 ):
        """this function was added so that you could cache the views more easily if you wanted to use the dbsprockets caching system"""
        
        if hiddenFields is None:
            hiddenFields = []
        if disabledFields is None:
            disabledFields=[]
        if requiredFields is None:
            requiredFields=[]
        if omittedFields is None:
            omittedFields=[]
        if additionalFields is None:
            additionalFields=[]
        if fieldAttrs is None:
            fieldAttrs={}
        if widgets is None:
            widgets={}
        if validators is None:
            validators={}

        helper = self._getHelper(model)
        metadata   = helper.getMetadata(model)
        identifier = helper.getIdentifier(model)
        provider   = helper.getProvider(model)

        viewConfig = viewConfigType(provider, identifier)
        viewConfig._action = action
        attrs = ['requiredFields', 'disabledFields', 'hiddenFields', 'omittedFields', 'additionalFields']
        for attr in attrs:
            tipe = eval(attr) #wouldn't be a hack unless I had a crafty eval, right?
            fields = getattr(viewConfig, attr)
            for field in tipe:
                if field not in fields:
                    fields.append(field)
        if formValidator:
            viewConfig.widgetValidator = formValidator
        if limitFields:
            viewConfig.limitFields = limitFields
        if widgetType:
            viewConfig.widgetType = widgetType
        viewConfig.checkIfUnique = checkIfUnique
        viewConfig.fieldValidators.update(validators)
        viewConfig.fieldWidgets.update(widgets)
        viewConfig.fieldAttrs.update(fieldAttrs)

        if makeLinks is not None:
            tipe = type(viewConfig)
            method = new.instancemethod(makeLinks, viewConfig, tipe)
            viewConfig._makeLinks = method
            
        view = ViewFactory().create(viewConfig)
        return view
    
    def makeTable(self, model, 
                  controller=None,
                  identifier='',
                  omittedFields=None, 
                  additionalFields=None, 
                  limitFields=None,
                  widgetType=None,
                  makeLinks=None,
                  ):
        """
        makeTable(model,
                 controller=None,
                 identifier='',
                 omittedFields=[],
                 additionalFields=[],
                 limitFields=None,
                 widetType=None,
                 makeLinks=None,
                 )-->Toscawidgets.DataGrid
                 
        Generates a datagrid widget for population of database data
        
        Arguments:
        name              type                           default    Description
        model             class                                     sqlalchemy mapped class definition
        controller        String                                    what controller to send the edit and delete links to.
        identifier        String or None                 None       optional identifier for the form
        omittedFields     list of names                  []         fields that are not present, whatsoever.
                                                                    if '_action' is in the list, then the edit/delete links will be removed.
        additionalFields  list of Toscawidgets.fields    []         fields that you would like to add to the form that are not in the model
        limitFields       list of names                  None       limit the fields to the one in the list.  A side effect is that the 
                                                                    order of the fields is set by this argument.  
                                                                    (a good way to order your fields if you include all of them)
        widgetType        ToscaWidget                    None       field for the parent widget to be used
        makeLinks         function                       None       Add a custom method for creating action links in your table.
        """
        if omittedFields is None:
            omittedFields=[]
        if additionalFields is None:
            additionalFields=[]
        
        self._validateModel(model)
        if controller is None:
            controller = ''
            omittedFields.append('_actions')

        view = self.makeView(TableViewConfig, 
                        model, 
                        controller=controller, 
                        omittedFields=omittedFields, 
                        additionalFields=additionalFields,
                        widgetType=widgetType,
                        makeLinks=makeLinks,
                        )
        return view.widget

class DataFetcher(DatabaseMixin):
        
    def getTableValue(self, model, offset=None, limit=None, foreignKeyFieldLabels={}):
        """
        getTableValue(model) --> SQLAlchemy result rows
        gets a set of results for a given model
        
        xxx: limit and offset are currently not employed
        
        Arguments:
        name              type                           Description
        model             class                          sqlalchemy mapped class definition
        """

        helper = self._getHelper(model)
        metadata = helper.getMetadata(model)
        identifier = helper.getIdentifier(model)
        provider = helper.getProvider(model)
        config = TableViewSessionConfig(identifier+'_session', provider, identifier=identifier)
        return config.getValue()

    def getFormDefaults(self, model):
        """
        getTableValue(model) --> SQLAlchemy result rows
        gets a set of results for a given model
        
        Arguments:
        name              type                           Description
        model             class                          sqlalchemy mapped class definition
        """

        helper = self._getHelper(model)
        identifier = helper.getIdentifier(model)
        provider = helper.getProvider(model)
        return provider.getDefaultValues(identifier)
 
generator = SprocketGenerator()
fetcher = DataFetcher()

makeForm = generator.makeForm
makeTable = generator.makeTable
makeRecordView = generator.makeRecordView
getTableValue = fetcher.getTableValue
getFormDefaults = fetcher.getFormDefaults



import os.path

import paste.fileapp
from pylons import request
from pylons.middleware import error_document_template, media_path

from tg2testapp.lib.base import BaseController

class ErrorController(BaseController):
    """Generates error documents as and when they are required.

    The ErrorDocuments middleware forwards to ErrorController when error
    related status codes are returned from the application.

    This behaviour can be altered by changing the parameters to the
    ErrorDocuments middleware in your config/middleware.py file.
    """

    def document(self):
        """Render the error document"""
        resp = request.environ.get('pylons.original_response')
        page = error_document_template % \
            dict(prefix=request.environ.get('SCRIPT_NAME', ''),
                 code=request.params.get('code', resp.status_int),
                 message=request.params.get('message', resp.body))
        return page

    def img(self, id):
        """Serve stock images"""
        return self._serve_file(os.path.join(media_path, 'img', id))

    def style(self, id):
        """Serve stock stylesheets"""
        return self._serve_file(os.path.join(media_path, 'style', id))

    def _serve_file(self, path):
        """Call Paste's FileApp (a WSGI application) to serve the file
        at the specified path
        """
        fapp = paste.fileapp.FileApp(path)
        return fapp(request.environ, self.start_response)



"""Pylons middleware initialization"""
from beaker.middleware import SessionMiddleware, CacheMiddleware
from paste.cascade import Cascade
from paste.registry import RegistryManager
from paste.urlparser import StaticURLParser
from paste.deploy.converters import asbool
from paste import httpexceptions

from pylons import config
from pylons.middleware import error_mapper, ErrorDocuments, ErrorHandler, \
    StaticJavascripts, StatusCodeRedirect
from pylons.wsgiapp import PylonsApp
from routes.middleware import RoutesMiddleware
from tw.core.middleware import TGWidgetsMiddleware
from tw.mods.pylonshf import PylonsHostFramework


from  tg2testapp.config.environment import load_environment

def make_app(global_conf, full_stack=True, **app_conf):
    """Create a Pylons WSGI application and return it

    ``global_conf``
        The inherited configuration for this application. Normally from
        the [DEFAULT] section of the Paste ini file.

    ``full_stack``
        Whether or not this application provides a full WSGI stack (by
        default, meaning it handles its own exceptions and errors).
        Disable full_stack when this application is "managed" by
        another WSGI middleware.

    ``app_conf``
        The application's local configuration. Normally specified in the
        [app:<name>] section of the Paste ini file (where <name>
        defaults to main).
    """
    # Configure the Pylons environment
    load_environment(global_conf, app_conf)
    from  tg2testapp.controllers.root import RootController

    # Create the TurboGears WSGI app stack
    app = PylonsApp()

    # Add Routes, Session and Cashing support
    app = RoutesMiddleware(app, config['routes.map'])
    app = SessionMiddleware(app, config)
    app = CacheMiddleware(app, config)

    # Configure and add ToscaWidgets support
    host_framework = PylonsHostFramework(default_view="genshi")
    app = TGWidgetsMiddleware(app, host_framework)

    # CUSTOM MIDDLEWARE HERE (filtered by error handling middlewares)

    if asbool(full_stack):
        # Handle Python exceptions
        app = ErrorHandler(app, global_conf, **config['pylons.errorware'])

        # Display error documents for 401, 403, 404 status codes (and
        # 500 when debug is disabled)
        if asbool(config['debug']):
            app = StatusCodeRedirect(app)
        else:
            app = StatusCodeRedirect(app, [400, 401, 403, 404, 500])

    # Establish the Registry for this application
    app = RegistryManager(app)

    # Static files (If running in production, and Apache or another web
    # server is handling this static content, remove the following 3 lines)
    javascripts_app = StaticJavascripts()
    static_app = StaticURLParser(config['pylons.paths']['static_files'])
    app = Cascade([static_app, javascripts_app, app])
    return app



from nose.tools import raises
from dbsprockets.validatorselector import ValidatorSelector, SAValidatorSelector, UniqueValue
from dbsprockets.test.base import *
from tw.forms.validators  import *
from formencode.compound import All
from sqlalchemy import Column, Integer
from sqlalchemy.databases.oracle import *
from dbsprockets.saprovider import SAProvider
from types import NoneType

def setup():
    setupDatabase()

provider = SAProvider(metadata)
class TestValidatorSelector:
    def setup(self):
        self.validatorSelector = ValidatorSelector()
        
    def testCreateObj(self):
        pass
    
    def testSelect(self):
        assert self.validatorSelector.select('lala') == UnicodeString

class TestSAValidatorSelector:
    testColumns = (
    (BLOB,        NoneType),
    (BOOLEAN,     NoneType),
    (Binary,      NoneType),
    (Boolean,     NoneType),
    (CHAR,        UnicodeString),
    (CLOB,        UnicodeString),
    (DATE,        DateValidator),
    (DATETIME,    DateValidator),
    (DECIMAL,     Number),
    (Date,        DateValidator),
    (DateTime,    DateValidator),
    (FLOAT,       Number),
    (Float,       Number),
    (INT,         Int),
    (Integer,     Int),
    (Numeric,     Number),
    (PickleType,  UnicodeString),
    (SMALLINT,    Int),
    (SmallInteger,Int),
    (String,      UnicodeString),
    (TEXT,        UnicodeString),
    (TIME,        DateValidator),
    (Time,        DateValidator),
    (TIMESTAMP,   DateValidator),
    (Unicode,     UnicodeString),
    (VARCHAR,     UnicodeString),
    (OracleNumeric,      Number),
    (OracleDate,         DateValidator),
    (OracleDateTime,     DateValidator),
    (OracleInteger,      Int),
    (OracleSmallInteger, Int),

    )

    def setup(self):
        self.validatorSelector = SAValidatorSelector(provider)
        
    def testCreateObj(self):
        pass
    
    def testSelect(self):
        for type, expected in self.testColumns:
            args={}
            if isinstance(type, Text):
                args['size'] = 100
            c = Column('asdf', type, args)
            yield self._testSelect, c, expected
            
    def _testSelect(self, column, expected):
        validator = self.validatorSelector.select(column)
        assert validator.__class__ == expected, "expected: %s\nactual: type: %s validator: %s"%(expected, column.type, validator.__type__)
        
    @raises(TypeError)
    def _select(self, arg1):
        self.validatorSelector.select(arg1)
    
    def testSelectBad(self):
        badInput = ('a', 1, {}, [], (), None, 1.2)
        for input in badInput:
            yield self._select, input
    
    def testNameBasedValidatorSelect(self):
        c = Column('email_address', String)
        validator = self.validatorSelector.select(c)
        assert isinstance(validator, Email)
        
    def testValidatorSelectorUniqueField(self):
        c = Column('nana', String, unique=True)
        validator= self.validatorSelector.select(c, checkIfUnique=True)
        assert type(validator) is All

class TestUniqueValue:
    
    def setup(self):
        self.validator = UniqueValue(SAProvider(metadata), users_table.c.user_name)
        
    @raises(Invalid)
    def testToPythonInvalid(self):
        self.validator.to_python(u'asdf', None)
        
    def testToPythonValid(self):
        self.validator.to_python(u'asdf1234', None)



#this might go away.

from view import View

class ViewFactory:
    def create(self, viewConfig, id=None):
        kw = viewConfig.getWidgetArgs()
        if id == None:
            id = viewConfig.__class__.__name__+'_'+viewConfig.identifier
        kw['id'] = id
        parentWidget = viewConfig.widgetType(**kw)
        return View(parentWidget, viewConfig)



from nose.tools import raises, eq_
from dbsprockets.saprovider import SAProvider
from dbsprockets.test.testIProvider import TestIProvider
from model import *
from cgi import FieldStorage
from dbsprockets.test.base import setupDatabase, sortedTableList, teardownDatabase
from cStringIO import StringIO

def setup():
    setupDatabase()
def teardown():
    teardownDatabase()

class _TestSAProvider(TestIProvider):
    def setUp(self):
        self.provider = SAProvider(metadata)
        
    def testCreate(self):
        pass
    
    @raises(TypeError)
    def _create(self, arg1):
        SAProvider(arg1)
        
    def testCreateBad(self):
        badInput = (1, 1.2, u'a', None, (), [], {})
        for input in badInput:
            yield self._create, badInput

    def testGetTables(self):
        tables = sorted(self.provider.getTables())
        eq_(tables, sortedTableList)

    def testGetTable(self):
        table = self.provider.getTable('tg_user')
        eq_(table.name, 'tg_user')

    def testGetColumns(self):
        columns = sorted(self.provider.getColumns('tg_user'))
        eq_(columns, ['created', 'display_name', 'email_address', 'password', 'town', 'user_id', 'user_name'])

    def testGetColumn(self):
        column = self.provider.getColumn('tg_user', 'user_id')
        eq_(column.name, 'user_id')

    def testGetPrimaryKeys(self):
        keys = self.provider.getPrimaryKeys('tg_user')
        eq_(keys, ['user_id'])

    def testSelectOnPks(self):
        rows = self.provider.selectOnPks('tg_user', dict(user_id=1))
        actual = (rows[0])[:-1]
        assert list(actual) == [1, u'asdf', None, None, 'asdf', 1], "%s"%list(actual)

    def testSelectOnPksWithColumnsLimit(self):
        rows = self.provider.selectOnPks('tg_user', dict(user_id=1), columnsLimit=['user_name'])
        actual = rows[0]
        assert list(actual) == [u'asdf', ], "%s"%list(actual)

    def testSelectWithLimitAndOffset(self):
        rows = self.provider.select('tg_group', dict(user_id=1), columnsLimit=['group_name'], resultLimit=10, resultOffset=2)
        actual = rows
        assert list(actual) == [(u'2',), (u'3',), (u'4',), (u'5',), (u'6',), (u'7',), (u'8',), (u'9',), (u'10',), (u'11',)], "%s"%list(actual)

    def testSelect(self):
        rows = self.provider.select('tg_user')
        actual = (rows[0])[:-1]
        assert list(actual) == [1, u'asdf', None, None, 'asdf', 1], "%s"%list(actual)
        
    def testSelectWithColumnsLimit(self):
        rows = self.provider.select('tg_user', columnsLimit=['user_name'])
        actual = rows[0]
        assert list(actual) == [u'asdf'], "%s"%list(actual)

    def testSelectEmpty(self):
        rows = self.provider.select('tg_user', values=dict(user_name=u'asdf', user_id=0))
        assert rows == [], "%s"%rows

    def testAdd(self):
        self.provider.add('tg_user', values=dict(user_name=u'asdf2', user_id=0))
        
    def testAddWithColumnsLimit(self):
        self.provider.add('tg_user', values=dict(user_name=u'asdf4', user_id=0, enabled=True), columnsLimit=['user_name', 'user_id'])
        
    def testDelete(self):
        self.provider.add('tg_user', values=dict(user_name=u'asdf3'))
        self.provider.delete('tg_user', dict(user_id=2))

    def testEdit(self):
        self.provider.edit('tg_user', values=dict(user_id=0, user_name=u'asdf2'))
        rows = users_table.select(users_table.c.user_name==u'asdf').execute().fetchall()
        assert len(rows) == 1
        row = rows[0][:-1]
        assert row == (1, u'asdf', None, None, u'asdf', 1), "%s"%row
        
    def testGetViewColumnName(self):
        actual = self.provider.getViewColumnName('town_table')
        assert actual == 'name', actual
        
    def testGetIDColumnName(self):
        actual = self.provider.getIDColumnName('town_table')
        assert actual == 'town_id', actual
        
    def testGetForeignKeys(self):
        actual = self.provider.getForeignKeys('tg_user')
        c = users_table.columns['town']
        assert actual == [c, ], actual
        
    def testGetAssociatedManyToManyTables(self):
        actual = self.provider.getAssociatedManyToManyTables('tg_user')
        assert actual == ['tg_group', ], actual
    
    def testGetManyToManyColumns2(self):
        actual = sorted(self.provider.getManyToManyColumns('tg_group'))
        eq_(actual,  [u'permissions', u'tg_users'])

    def testGetManyToManyColumns(self):
        actual = self.provider.getManyToManyColumns('tg_user')
        eq_(actual,   [u'tg_groups'])

    def testGetManyToManyTables(self):
        actual = sorted(self.provider.getManyToManyTables())
        assert actual == ['group_permission', 'user_group'], "%s"%actual
        
    def testSetManyToMany(self):
        self.provider.setManyToMany('tg_user', 1, 'tg_group', [1,2])
        rows = user_group_table.select().execute().fetchall()
        assert rows == [(1, 1), (1, 2)], "%s"%rows

    def testSetManyToManyToo(self):
        user_group_table.delete().execute()
        self.provider.setManyToMany('tg_group', 1, 'tg_user', [1,])
        rows = user_group_table.select().execute().fetchall()
        assert rows == [(1, 1),], "%s"%rows
        
    def testSetManyToManyFailsafe(self):
        r = self.provider.setManyToMany('tg_group', 1, 'test_table', [1,])
        eq_(r, None)
    
    def testGetForeignKeyDict(self):
        d = self.provider.getForeignKeyDict('tg_user')
        assert d == {'town': {1: u'Arvada', 2: u'Denver', 3: u'Golden', 4: u'Boulder'}},  "%s"%d

    def test_findFirstColumnColumnNotFound(self):
        actual = self.provider._findFirstColumn('test_table', ['not_possible',])
        eq_(actual, 'id')
        
    def testIsUniqueNotUnique(self):
        actual = self.provider.isUnique(users_table.c.user_name, u'asdf')
        assert not actual

    def testIsUnique(self):
        actual = self.provider.isUnique(users_table.c.user_name, u'asdf_q2341234')
        assert actual
        
    def testAddFieldStorage(self):
        self.provider.add('test_table', values=dict(BLOB=FieldStorage('asdf.txt', StringIO())))
        
    def testGetDefaultValues(self):
        #we dont want to check this for reflected tables (there are no defaults set):
        if 'Reflected' in self.__class__.__name__:
            return
        actual = self.provider.getDefaultValues('test_table')
        assert sorted(actual.keys()) == ['Integer', 'created'], sorted(actual.keys())
        assert actual['Integer'] == 10
        
class TestSAProvider(_TestSAProvider):
    pass





from nose.tools import raises
from tw.api import Widget

from dbsprockets.iprovider import IProvider
from dbsprockets.sprockets import Views, Sessions, Sprocket, Sprockets, SessionConfigError, ViewConfigError
from dbsprockets.view import View
from dbsprockets.sessionconfig import SessionConfig
from dbsprockets.viewconfig import ViewConfig
from dbsprockets.metadata import Metadata

class TestSprockets:
    pass

class TestViews:
    provider = IProvider()
    def setup(self):
        self.views = Views(self.provider, '')
        
    def testCreate(self):
        pass
    
    @raises(TypeError)
    def _create(self, provider, controller=''):
        Views(provider, controller)
        
    def testCreateBad(self):
        provider = IProvider()
        badInput = ('a', (), {}, [], 1, 1.2)
        for input in badInput:
            yield self._create, badInput
        for input in badInput[1:]:
            yield self._create, provider, badInput
            
    @raises(TypeError)
    def testPutBad(self):
        self.views['lala'] = 'lala'
    
    def testPut(self):
        viewConfig = ViewConfig(self.provider,'')
        view = View(Widget, viewConfig)
        self.views['test'] = view
        
    @raises(ViewConfigError)
    def testGetBadViewConfigType(self):
        self.views['testBad__tableName']

    def testGet(self):
        viewConfig = ViewConfig(self.provider,'')
        view = View(Widget, viewConfig)
        self.views['test'] = view
        assert self.views['test'] == view
        
    @raises(TypeError)
    def testGetBad(self):
        self.views[1]
        
class TestSessions:
    provider = IProvider()
    def setup(self):
        self.sessions = Sessions(self.provider)
        
    def testCreate(self):
        pass
    
    @raises(TypeError)
    def _create(self, arg1):
        Sessions(arg1)
        
    def testCreateBad(self):
        badInput = ((), {}, [], 'a', 1, 1.2)
        for input in badInput:
            yield self._create, badInput
            
    @raises(TypeError)
    def testPutBad(self):
        self.sessions['lala'] = 'lala'
    
    def testPut(self):
        sessionConfig = SessionConfig('', self.provider)
        self.sessions['test'] = sessionConfig
    
    def testGet(self):
        sessionConfig = SessionConfig('', self.provider)
        self.sessions['test'] = sessionConfig
        assert self.sessions['test'] is sessionConfig
    
    @raises(SessionConfigError)
    def testGetBadSessionConfigType(self):
        self.sessions['testBad__tableName']
        
    @raises(TypeError)
    def testGetBad(self):
        self.sessions[1]
        
class TestSprocket:
    provider = IProvider()
    sessionConfig = SessionConfig('',provider)
    viewConfig = ViewConfig(provider, '')
    view = View(Widget, viewConfig)
    
    def setup(self):
        self.sprocket = Sprocket(self.view, self.sessionConfig)
        
    def testCreate(self):
        pass
    
    @raises(TypeError)
    def _create(self, arg1, arg2=None):
        Sprocket(arg1, arg2)
        
    def testCreateBad(self):
        badInput = ((), {}, [], 'a', 1, 1.2)
        for input in badInput:
            yield self._create, input

        badInput = ((), {}, [], 'a', 1, 1.2)
        for input in badInput:
            yield self._create, self.view, input

class TestSprockets:
    provider = IProvider()
    def setup(self):
        self.sprockets = Sprockets(self.provider, '')
        
    def testCreate(self):
        pass
    
    @raises(TypeError)
    def _create(self, arg1, arg2=''):
        Sprockets(arg1, arg2)
        
    def testCreateBad(self):
        provider = IProvider()
        badInput = ('a', (), {}, [], 1, 1.2)
        for input in badInput:
            yield self._create, badInput
        for input in badInput[1:]:
            yield self._create, provider, badInput
            
    @raises(TypeError)
    def testPutBad(self):
        self.sprockets['lala'] = 'lala'
    
    def testPut(self):
        sessionConfig = SessionConfig('', self.provider)
        viewConfig = ViewConfig(self.provider, '')
        view = View(Widget, viewConfig)
        sprocket = Sprocket(view, sessionConfig)
        self.sprockets['test'] = sprocket
        
    def testGet(self):
        sessionConfig = SessionConfig('',self.provider)
        viewConfig = ViewConfig(self.provider, '')
        view = View(Widget, viewConfig)
        sprocket = Sprocket(view, sessionConfig)
        self.sprockets['test'] = sprocket
        gotSprocket = self.sprockets['test']
        assert gotSprocket == sprocket
        
    @raises(TypeError)
    def testGetBad(self):
        self.sprockets[1]


"""
ViewConfig Module

The ViewConfig defines all of the View related issues for dbsprockets

Classes:
Name                               Description
ViewConfig                         Parent Class
DatabaseViewConfig
EditRecordConfig
TableDefViewConfig
TableViewConfig
AddRecordViewConfig

Exceptions:
None

Functions:
None

Copywrite (c) 2007 Christopher Perkins
Original Version by Christopher Perkins 2007
Released under MIT license.
"""
import types
from cStringIO import StringIO
from tw.forms.fields import HiddenField
from tw.api import Widget
from tw.forms import DataGrid, TableForm, SingleSelectField
from copy import copy
from urllib import quote
from genshi.input import HTML, XML 

from dbsprockets.validatorselector import ValidatorSelector, SAValidatorSelector
from dbsprockets.widgetselector import *
from dbsprockets.widgets.widgets import *
from dbsprockets.metadata import *
from dbsprockets.iprovider import IProvider
from formencode.validators import String
from formencode.compound import All
from formencode import Schema

class FilteringSchema(Schema):
    filter_extra_fields = True
    allow_extra_fields = True

from tw.forms.fields import SingleSelectField

class ViewConfigMeta(object):
    _hiddenFields     = []
    _disabledFields   = []
    _requiredFields   = []
    _omittedFields    = []
    _additionalFields = []
    
    limitFields = None

    validatorSelectorType = ValidatorSelector
    widgetValidator  = None
    checkIfUnique  = False

    _fieldWidgets     = {}
    _fieldAttrs       = {}
    _fieldValidators  = {}

    widgetType = Widget
    widgetSelectorType = WidgetSelector

    metadataType = Metadata

    def __new__(cls, *args, **kw):
        bases = cls.mro()
        chainedAttrs = ['_hiddenFields', '_requiredFields', '_disabledFields', '_omittedFields']
        for base in bases:
            for attr in chainedAttrs:
                if hasattr(base, attr):
                    current =  getattr(cls, attr)
                    setattr(cls, attr, current + [item for item in getattr(base, attr) if item not in current])
        return object.__new__(cls, *args, **kw)

    def __init__(self, provider, identifier=None, controller='/'):
        """(metadata, controller)"""
        if not isinstance(provider, IProvider):
            raise TypeError('arg1 should be an IProvider, not a : %s', type(provider))
        if not isinstance(identifier, types.StringTypes):
            raise TypeError('arg2 should be a String, not a: %s', type(identifier))
        if not isinstance(controller, types.StringTypes):
            raise TypeError('arg3 should be a String,  not a: %s', type(controller))
        
        self.provider = provider
        self.identifier = identifier
        self.controller = controller
        self.metadata = self.metadataType(provider, identifier)
        self.widgetSelector = self._doGetWidgetSelector()
        self.validatorSelector = self.validatorSelectorType(provider)
        
        self.__hiddenFields     = copy(self._hiddenFields)
        self.__disabledFields   = copy(self._disabledFields)
        self.__requiredFields   = copy(self._requiredFields)
        self.__omittedFields    = copy(self._omittedFields)
        self.__additionalFields = copy(self._additionalFields)
     
        self.fieldWidgets       = copy(self._fieldWidgets)
        self.fieldAttrs         = copy(self._fieldAttrs)
        self.fieldValidators    = copy(self._fieldValidators)

    def _getOmittedFields(self):
        return self.__omittedFields
    def _setOmittedFields(self, item):
        self.__omittedFields = item
    omittedFields = property(_getOmittedFields, _setOmittedFields)

    def _getAdditionalFields(self):
        return self.__additionalFields
    def _setAdditionalFields(self, item):
        self.__additionalFields = item
    additionalFields = property(_getAdditionalFields, _setAdditionalFields)

    def _getDisabledFields(self):
        fields =  self.__disabledFields
        fields.extend([field for field in self._doGetDisabledFields() if field not in fields])
        return fields
    def _setDisabledFields(self, item):
        self.__disabledFields = item
    disabledFields = property(_getDisabledFields, _setDisabledFields)

    def _getRequiredFields(self):
        return self.__requiredFields
    def _setRequiredFields(self, item):
        self.__requiredFields = item
    requiredFields = property(_getRequiredFields, _setRequiredFields)

    def _getHiddenFields(self):
        return self.__hiddenFields
    def _setHiddenFields(self, item):
        self.__hiddenFields = item
    hiddenFields = property(_getHiddenFields, _setHiddenFields)

    @property
    def manytoManyFields(self):
        return self._doGetManyToManyFields()

    @property
    def fields(self):
        if self.limitFields is not None:
            return self.limitFields
        return self.metadata.keys()

    def _createHiddenFields(self):
        fields = []
        idField = HiddenField(id='dbsprockets_id')
        fields.extend([HiddenField(id=field, identifier=field) for field in self.hiddenFields if field not in self.omittedFields]  )
        fields.extend([HiddenField(id=field, identifier=field) for field in self.disabledFields if field not in self.omittedFields and field not in self.hiddenFields])
        fields.append(idField)
        return fields
    
    def getWidgetArgs(self, id=None):
        fieldAttrs = self.fieldAttrs
        hiddenFields = self.hiddenFields
        disabledFields = self.disabledFields
        requiredFields = self.requiredFields
        omittedFields = self.omittedFields
        validators = self.fieldValidators
        fieldWidgets = self.fieldWidgets
        fields = self.fields
        manyToManyFields = self.manytoManyFields

        children = []
        for fieldName in fields:
            if fieldName in fieldWidgets:
                children.append(fieldWidgets[fieldName])
                continue
            if fieldName in omittedFields:
                continue
            if fieldName in manyToManyFields:
                continue
            
            field = self.metadata[fieldName]
            widgetType = self.widgetSelector.select(field)
            childWidgetArgs = self._doGetChildWidgetArgsForWidgetType(field, widgetType)
            childWidgetArgs.update({'id':fieldName, 'identifier':field})
            
            required = fieldName in requiredFields
            disabled = fieldName in disabledFields
            hidden   = fieldName in hiddenFields
            checkIfUnique = self.checkIfUnique

            validator = None
            if not disabled:
                if fieldName in validators:
                    validator = validators[fieldName]
                else:
                    validator = self.validatorSelector.select(field, required, checkIfUnique)
            if validator != None:
                childWidgetArgs['validator'] = validator

            if fieldName in disabledFields:
                childWidgetArgs['disabled'] = True
                childWidgetArgs['id'] = fieldName
                childWidgetArgs['validator'] = String(if_missing=None)

            if fieldName in fieldAttrs:
                childWidgetArgs['attrs'] = fieldAttrs[fieldName]
                
            children.append(widgetType(**childWidgetArgs))
        children.extend(self._createHiddenFields())
        children.extend(self._doGetManyToManyWidgets(manyToManyFields))
        children.extend(self.additionalFields)
        d = dict(children=children)
        d['controller'] = self.controller
        if self.widgetValidator:
            d['validator'] = self.widgetValidator
        else:
            d['validator'] = FilteringSchema
        d.update(self._doAddExtraWidgetArgs())
        return d

    @property
    def action(self):
        return self._doGetAction()
    
    def _doGetManyToManyFields(self):
        return []
    def _doGetAction(self):
        raise NotImplementedError
    def _doAddExtraWidgetArgs(self):
        return {}
    def _doGetWidgetSelector(self):
        return self.widgetSelectorType()
    def _doGetChildWidgetArgsForWidgetType(self, field, widgetType):
        return {}
    def _doGetManyToManyWidgets(self, fields):
        return []
    def _doGetDisabledFields(self):
        return []
    
class ViewConfig(ViewConfigMeta):pass

class DatabaseViewConfig(ViewConfig):
    widgetType = ContainerWidget
    metadataType = DatabaseMetadata
    widgetSelectorType = DatabaseViewWidgetSelector

class TableDefViewConfig(ViewConfig):
    widgetType = TableWidget
    metadataType = FieldsMetadata
    widgetSelectorType = TableDefWidgetSelector

class RecordViewConfig(ViewConfig):
    widgetType = RecordViewWidget
    metadataType = FieldsMetadata
    widgetSelectorType = RecordViewWidgetSelector

class TableViewConfig(ViewConfig):
    widgetType = DataGrid
    metadataType = FieldsMetadata
    foreignKeyFieldLabels = ['name', '_name', 'description', '_description', 'title']

    def _writePKsToURL(self, pks, row):
        stream = StringIO()
        l = len(pks) - 1
        for i, pk in enumerate(pks):
            if i == 0:
                stream.write('?')
            else:
                stream.write('&')
            stream.write(str(pk))
            stream.write('=')
            value=row[pk]
            if hasattr(value,"original"):
                value=value.original
            if isinstance(value, unicode):
                value=value.encode('utf8')
            value=quote(str(value))
            stream.write(value)
        
        return stream.getvalue()
    
    def _makeLinks(self, row):
        pks = self.metadata.primaryKeys()
        editLink = StringIO()
        editLink.write('<a href="')
        editLink.write(self.controller+'/editRecord/')
        editLink.write(self.metadata.identifier)
        editLink.write(self._writePKsToURL(pks, row))
        editLink.write('">edit</a>')

        deleteLink = StringIO()
        deleteLink.write('<a href="')
        deleteLink.write(self.controller+'/delete/')
        deleteLink.write(self.metadata.identifier)
        deleteLink.write(self._writePKsToURL(pks, row))
        deleteLink.write('">delete</a>')
        editLink.write('|')
        editLink.write(deleteLink.getvalue())
        return HTML(editLink.getvalue())

    def getWidgetArgs(self):
        fields = self.fields
        pks  = self.metadata.primaryKeys()
        omittedFields = self.omittedFields
        fields = [(field, eval('lambda d: d["'+field+'"]')) for field in fields if field not in omittedFields]
        if '_actions' not in omittedFields:
            fields.insert(0, (' ', self._makeLinks))
        manyToManyTables = self.provider.getManyToManyTables()
        if self.identifier not in manyToManyTables:
            manyToManyColumns = self.provider.getManyToManyColumns(self.identifier)
            fields.extend([('%s'%table, eval('lambda d: d["'+table+'"]')) for table in manyToManyColumns])
        return dict(id=self.__class__.__name__, fields=fields)
        
class EditableRecordViewConfig(ViewConfig):
    widgetType = DBSprocketsTableForm
    metadataType = FieldsMetadata
    widgetSelectorType = SAWidgetSelector
    validatorSelectorType = SAValidatorSelector
    _hiddenFields     = ['tableName']
    _action = ''
    def _doGetAction(self):
        return self._action

    def _doAddExtraWidgetArgs(self):
        return dict(action=self.action)

    def _doGetDisabledFields(self):
        return self.metadata.primaryKeys()
    
    def _doGetManyToManyFields(self):
        associatedTables = self.provider.getAssociatedManyToManyTables(self.identifier)
        return [table+'s' for table in associatedTables]
    
    def _doGetManyToManyWidgets(self, fields):
        tableName = self.identifier
        widgets = []
        omittedFields = self.omittedFields
        hiddenFields  = self.hiddenFields
        limitFields   = self.limitFields
        for fieldName in fields:
            table = fieldName[:-1]
            if not (fieldName in hiddenFields ) and \
               not (fieldName in omittedFields) and \
                   ((limitFields is None) or (fieldName in limitFields)):
                        widgets.append(ForeignKeyMultipleSelectField(id="many_many_"+table, label_text=fieldName, tableName=table, provider=self.provider))
        return widgets
    
    def _doGetChildWidgetArgsForWidgetType(self, field, widgetType):
        if widgetType == ForeignKeySingleSelectField:
            nullable=self.provider.isNullableField(field)
            return dict(nullable=nullable, validator=String(if_missing=None), provider=self.provider, tableName=field.foreign_keys[0].column.table.name)
        return {}

class EditRecordViewConfig(EditableRecordViewConfig):
    widgetType = DBSprocketsTableForm
    metadataType = FieldsMetadata
    widgetSelectorType = SAWidgetSelector
    validatorSelectorType = SAValidatorSelector
    _hiddenFields     = ['tableName']
    
    def _doGetAction(self):
        return self.controller+'/edit'

class AddRecordViewConfig(EditRecordViewConfig):
    checkIfUnique = True
    def _doGetAction(self):
        return self.controller+'/add'

    def _doGetDisabledFields(self):
        foreignKeys=self.metadata.foreignKeys
        return [k for k in self.metadata.primaryKeys() if k in self.metadata.autoIncrementFields and not k in foreignKeys]


#setup.py
from setuptools import setup, find_packages

setup(
  name="DBSprockets",
  version="0.2",
  zip_safe=False,
  include_package_data=True,
  description="""A package for creation of web widgets directly from database definitions.""",
  author="Christopher Perkins 2007",
  author_email="chris@percious.com",
  license="MIT",
  url="http://dbsprockets.googlecode.com/svn/dist",
  install_requires=['sqlalchemy>=0.4',
                    'tw.forms>0.8',
                    'Genshi>=0.4',
                    ],
  packages = find_packages(),
  classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
 entry_points = """[toscawidgets.widgets]
        # Use 'widgets' to point to the module where widgets should be imported
        # from to register in the widget browser
        widgets = dbsprockets.widgets
        # Use 'samples' to point to the module where widget examples
        # should be imported from to register in the widget browser
        # samples = tw.samples
        # Use 'resources' to point to the module where resources
        # should be imported from to register in the widget browser
        #resources = dbsprockets.widgets.resources
       """
#  entry_points="""
#        [paste.paster_create_template]
#        dbsprockets=dbsprockets.instance.newDBSprockets:Template
#    """,

  )



"""
Sprockets Module

This is sort of like the central nervous system of dbsprockets.  Views and
Sessions are collected in separate caches and served up as sprockets.
The cache objects may be solidified at some point with a parent class.
They work for right now.


Classes:
Name           Description
Sprockets      A cache of Sprockets
Sprocket       A binding of Session and View configs
Views          A cache of Views
Sessions       A cache of Sessions

Exceptions:
SessionConfigError
ViewConfigError

Functions:
None

Copywrite (c) 2007 Christopher Perkins
Original Version by Christopher Perkins 2007
Released under MIT license.
"""
import types

from dbsprockets.viewfactory import ViewFactory
from dbsprockets.viewconfig import *
from dbsprockets.sessionconfig import *
from dbsprockets.saprovider import SAProvider
from dbsprockets.iprovider import IProvider
from dbsprockets.metadata import *
from dbsprockets.view import View

class SessionConfigError(Exception):pass
class ViewConfigError(Exception):pass

class Views(dict):
    def __init__(self, provider, controller):
        if not isinstance(provider, IProvider):
            raise TypeError('provider is not of type IProvider')
        if not isinstance(controller, types.StringTypes):
            raise TypeError('controller is not of type String')
        self.viewFactory = ViewFactory()
        self.provider = provider
        self.controller = controller
        
    defaultViewConfigs = {'databaseView': DatabaseViewConfig,
                          'editRecord'  : EditRecordViewConfig,
                          'addRecord'   : AddRecordViewConfig,
                          #'viewRecord'  : ViewConfig,
                          'tableView'   : TableViewConfig,
                          'tableDef'    : TableDefViewConfig,
                          }

    def _getView(self, key):
        if '__' not in key:
            identifier = ''
            viewType = key
        else:
            viewType, identifier = key.split('__')
        if viewType not in self.defaultViewConfigs:
            raise ViewConfigError('viewType:%s now found in default Views'%viewType)
        viewConfig = self.defaultViewConfigs[viewType](self.provider, identifier, self.controller)
        self[key] = self.viewFactory.create(viewConfig, id=key)
        return self[key]
    
    def __getitem__(self, key):
        if dict.__contains__(self, key):
            return dict.__getitem__(self, key)
        view = self._getView(key)
        self[key] = view
        return view

    def __setitem__(self, key, item):
        if not isinstance(item, View):
            raise TypeError('item must be of type View')
        return dict.__setitem__(self, key, item)
    
class Sessions(dict):
    def __init__(self, provider):
        if not isinstance(provider, IProvider):
            raise TypeError('provider is not of type IProvider')
        self.provider = provider

    defaultSessionConfigs = { 'databaseView': SessionConfig,
                              'tableDef'    : SessionConfig,
                              'viewRecord'  : SessionConfig,
                              'tableView'   : TableViewSessionConfig,
                              'editRecord'  : EditRecordSessionConfig,
                              'addRecord'   : AddRecordSessionConfig,
                             }

    def __getitem__(self, key):
        if dict.__contains__(self, key):
            return dict.__getitem__(self, key)
        session = self._getSession(key)
        self[key] = session
        return session

    def _getSession(self, key):
        if '__' not in key:
            identifier = ''
            viewType = key
        else:
            viewType, identifier = key.split('__')
        if viewType in self.defaultSessionConfigs:
            return self.defaultSessionConfigs[viewType](key, self.provider, identifier)
        raise SessionConfigError('Unknown session type')

    def __setitem__(self, key, item):
        if not isinstance(item, SessionConfig):
            raise TypeError('item must be of type SessionConfig')
        return dict.__setitem__(self, key, item)

class Sprocket:
    def __init__(self, view, session):
            
        if not isinstance(view, View):
            raise TypeError('view is not of type View')
        if not isinstance(session, SessionConfig):
            raise TypeError('session is not of type SessionConfig')
        self.session = session
        self.view = view

class Sprockets(dict):
    def __init__(self, provider, controller):
        if not isinstance(provider, IProvider):
            raise TypeError('provider is not of type IProvider')
        if not isinstance(controller, types.StringTypes):
            raise TypeError('controller is not of type String')
        self.views = Views(provider, controller)
        self.sessions = Sessions(provider)
        self.controller = controller

    def __getitem__(self, key):
        if key in self:
            return dict.__getitem__(self, key)
        sprocket = self.getSprocket(key)
        self[key] = sprocket
        return sprocket

    def getSprocket(self, key):
        view = self.views[key]
        session = self.sessions[key]
        return Sprocket(view, session)

    def __setitem__(self, key, item):
        if not isinstance(item, Sprocket):
            raise TypeError('item must be of type Sprocket')
        return dict.__setitem__(self, key, item)



"""
iprovider Module

this contains the class which allows dbsprockets to interface with any database.

Classes:
Name             Description
IProvider        provider interface

Exceptions:
None

Functions:
None


Copywrite (c) 2007 Christopher Perkins
Original Version by Christopher Perkins 2007
Released under MIT license.
"""

class IProvider:
    #table information stuff
    def getTables(self):
        """()->[list of tablesNames] 
        get a list of tables from the database"""
        raise NotImplementedError
    def getTable(self, name):
        """(name) -> sqlalchemy.schema.Table
        get the table definition with the given table name
        """
        raise NotImplementedError
    def getColumns(self, tableName):
        """(name) -> [list of columnNames]
        get the column names given the table name
        """
        raise NotImplementedError
    def getColumn(self, tableName, name):
        """(tableName, name) -> sqlalchemy.schema.Column
        get the column definition givcen the tablename and column name.
        """
        raise NotImplementedError
    def getPrimaryKeys(self, tableName):
        raise NotImplementedError

    #crud stuff
    def select(self, tableName, **kw):
        raise NotImplementedError
    def add(self, tableName=None, **kw):
        raise NotImplementedError
    def edit(self, tableName=None, **kw):
        raise NotImplementedError
    def isNullableField(self,field):
        raise NotImplementedError